# Manifest sections
MANIFEST_APPLICATION_VERSIONS = "ApplicationVersions"
MANIFEST_CONFIGURATION_ITEMS = "ConfigurationItems"
MANIFEST_ENVIRONMENT_DEFINITIONS = "EnvironmentDefinitions"

# Application tag attributes
MANIFEST_APPLICATION_KEY = "ApplicationKey"
MANIFEST_APPLICATION_NAME = "ApplicationName"
MANIFEST_APPLICATION_VERSION_KEY = "VersionKey"
MANIFEST_APPLICATION_VERSION_NUMBER = "VersionNumber"
MANIFEST_FLAG_IS_TEST_APPLICATION = "IsTestApplication"

# Configuration item attributes
MANIFEST_MODULE_KEY = "ModuleKey"
MANIFEST_MODULE_NAME = "ModuleName"
MANIFEST_CONFIG_ITEM_KEY = "ConfigurationItemKey"
MANIFEST_CONFIG_ITEM_NAME = "ConfigurationItemName"
MANIFEST_CONFIG_ITEM_TYPE = "ConfigurationItemType"
MANIFEST_CONFIG_ITEM_VALUES = "Values"
MANIFEST_CONFIG_ITEM_TARGET_VALUE = "TargetValue"

# Environment definitions attributes
MANIFEST_ENVIRONMENT_KEY = "EnvironmentKey"
MANIFEST_ENVIRONMENT_NAME = "EnvironmentName"
MANIFEST_ENVIRONMENT_LABEL = "EnvironmentLabel"

# Manifest attributes
MANIFEST_DEPLOYMENT_NOTES = "DeploymentNotes"

# AutoGen Manifest
MANIFEST_FOLDER = "trigger_manifest"
MANIFEST_FILE = "auto_gen_trigger_manifest.json"



# Python Modules
from toposort import toposort_flatten, CircularDependencyError

# Custom Modules
# Functions
from outsystems.cicd_probe.cicd_base import send_probe_get_request
# Variables
from outsystems.vars.cicd_vars import GET_APPLICATION_DEPENDENCIES_ENDPOINT, PROBE_DEPENDENCIES_SUCCESS_CODE


# Get a set of applications which are producers for a specified application version.
def get_app_dependencies(artifact_dir: str, probe_endpoint: str, api_key: str, application_version_key: str,
                         application_name: str, application_version: str):
    # Builds the API params
    params = {"ApplicationName": application_name, "ApplicationVersion": application_version}

    # Sends the request
    response = send_probe_get_request(
        probe_endpoint, GET_APPLICATION_DEPENDENCIES_ENDPOINT, api_key, params)
    status_code = response["http_status"]

    if status_code == PROBE_DEPENDENCIES_SUCCESS_CODE:
        response = response["response"]
        dependencies_list = []
        for dependency in response:
            dependencies_list.append(dependency["ApplicationKey"])
        return set(dependencies_list)
    else:
        raise NotImplementedError(
            "There was an error. Response from server: {}".format(response))


# Topological ordering (linear ordering) of a dependency list
def sort_app_dependencies(dep_list: list):
    try:
        return toposort_flatten(dep_list)
    except:
        raise CircularDependencyError(
            "There are circular dependencies among the list of applications.")



# Python Modules
import requests

# Custom Modules
from outsystems.exceptions.invalid_json_response import InvalidJsonResponseError
# Variables
from outsystems.vars.cicd_vars import PROBE_API_SSL_CERT_VERIFY
# Functions
from outsystems.vars.vars_base import get_configuration_value


# Method that builds the CICD Probe endpoint based on the environment host
def build_probe_endpoint(probe_http_proto: str, probe_url: str, probe_api_endpoint: str, probe_api_version: int):
    return "{}://{}/{}/v{}".format(probe_http_proto, probe_url, probe_api_endpoint, probe_api_version)


# Sends a GET request to LT, with url_params
def send_probe_get_request(probe_api: str, probe_endpoint: str, api_key: str, url_params: str):
    # Format the request URL to include the api endpoint
    request_string = "{}/{}".format(probe_api, probe_endpoint)
    # Set API key header, when provided
    headers = {"X-CICDProbe-Key": api_key} if api_key else None
    # Send the request
    response = requests.get(request_string, params=url_params, headers=headers, verify=get_configuration_value("PROBE_API_SSL_CERT_VERIFY", PROBE_API_SSL_CERT_VERIFY))
    response_obj = {"http_status": response.status_code, "response": {}}
    if len(response.text) > 0:
        try:
            response_obj["response"] = response.json()
        except:
            raise InvalidJsonResponseError(
                "GET {}: The JSON response could not be parsed. Response: {}".format(request_string, response.text))

    return response_obj



# Python Modules
import sys
import os
import argparse
import re
from time import sleep
import xml.etree.ElementTree as ET
from zipfile import ZipFile
from io import BytesIO

# Workaround for Jenkins:
# Set the path to include the outsystems module
# Jenkins exposes the workspace directory through env.
if "WORKSPACE" in os.environ:
    sys.path.append(os.environ['WORKSPACE'])
else:  # Else just add the project dir
    sys.path.append(os.getcwd())

# Custom Modules
# Variables
from outsystems.vars.file_vars import ARTIFACT_FOLDER, ENVIRONMENT_SOURCECODE_FOLDER, ENVIRONMENT_SOURCECODE_DOWNLOAD_FILE
from outsystems.vars.lifetime_vars import LIFETIME_HTTP_PROTO, LIFETIME_API_ENDPOINT, LIFETIME_API_VERSION
from outsystems.vars.manifest_vars import MANIFEST_APPLICATION_VERSIONS, MANIFEST_FLAG_IS_TEST_APPLICATION, \
    MANIFEST_APPLICATION_NAME
from outsystems.vars.pipeline_vars import SOURCECODE_SLEEP_PERIOD_IN_SECS, SOURCECODE_TIMEOUT_IN_SECS, SOURCECODE_ONGOING_STATUS, \
    SOURCECODE_FINISHED_STATUS
from outsystems.vars.dotnet_vars import MS_BUILD_NAMESPACE, ASSEMBLY_BLACKLIST

# Functions
from outsystems.lifetime.lifetime_base import build_lt_endpoint
from outsystems.lifetime.lifetime_environments import get_environment_app_source_code, get_environment_app_source_code_status, \
    get_environment_app_source_code_link, get_environment_key
from outsystems.lifetime.lifetime_applications import get_running_app_version
from outsystems.lifetime.lifetime_downloads import download_package
from outsystems.file_helpers.file import load_data
from outsystems.vars.vars_base import get_configuration_value, load_configuration_file

# ############################################################# SCRIPT ##############################################################


# Extract content of downloaded source code package (one folder per application module)
def extract_package_content(file_path: str, include_all_refs: bool, remove_resources_files: bool):
    module_count = 0
    with ZipFile(file_path, 'r') as zf:
        # Iterate through the content of the source code package
        for archive_name in zf.namelist():
            match = re.search(r'(.*)\.v\d+.zip$', archive_name)
            # Each package will have one .zip file per module
            if match:
                module_name = match.group(1)
                module_folder = os.path.join(os.path.dirname(file_path), "modules", module_name)
                file_data = BytesIO(zf.read(archive_name))
                with ZipFile(file_data) as zf2:
                    # Extract generated source code of each module to a subfolder
                    zf2.extractall(path=module_folder)

                # Check if any post-processing action is needed over the extracted resources
                if (include_all_refs or remove_resources_files):
                    process_csproj_files(module_name, module_folder, include_all_refs, remove_resources_files)

                # Update package modules count
                module_count += 1

    # Return number of modules inside the source code package
    return module_count


# Return the list of .csproj relative paths referenced in the module .sln file
def find_csproj_files(module_name: str, module_folder: str):

    # Builds the solution full path
    sln_file = os.path.join(module_folder, "{}.sln".format(module_name))

    # Final list of csprojs relative path found in the solution file
    csprojs = []

    # Read module solution file
    with open(sln_file, 'rb') as f:
        line = f.readline()
        while line:
            line = f.readline().decode('utf-8')
            if line.startswith("Project"):
                # Gets line's second object (i.e: csproj relative path)
                # Trim spaces and double quotes
                csproj = line.split(",")[1].strip().strip('\"')
                # Check if module's main csproj
                is_main = csproj == "{}.csproj".format(module_name)
                # Check if module's referencesProxy csproj
                is_referencesProxy = csproj.startswith("referencesProxy")
                # Append csproj details
                csprojs.append({"RelativePath": csproj, "IsMain": is_main, "IsReferencesProxy": is_referencesProxy})

    return csprojs


# Post-processing of .csproj files according to provided flags:
# --include_all_refs: Ensures that references proxy .csproj file reference all assemblies (.dll) available in the module bin folder
# --remove_resources_files: Removes embedded .resources files from main .csproj file (if existing)
def process_csproj_files(module_name: str, module_folder: str, include_all_refs: bool, remove_resources_files: bool):

    # Find .csproj files available in the provided module folder
    csprojs = find_csproj_files(module_name, module_folder)

    # Build bin directory full path
    bin_folder = os.path.join(module_folder, "bin")

    # Iterate through all csproj files for the current module
    for csproj in csprojs:
        # Build csproj file full path
        csproj_file = os.path.join(module_folder, csproj["RelativePath"])

        # Check if csproj is for the module's references proxy assembly
        if csproj["IsReferencesProxy"] and include_all_refs:
            # Read csproj file and identify first ItemGroup element
            ET.register_namespace('', MS_BUILD_NAMESPACE)
            tree = ET.parse(csproj_file)
            itemgroup_elem = tree.find("./{val}ItemGroup".format(val='{' + MS_BUILD_NAMESPACE + '}'))

            # Iterate through all dlls found in the bin directory
            for file in os.listdir(bin_folder):
                if file.endswith(".dll"):
                    dll_name = os.path.splitext(file)[0]
                    # TODO: Use os.path.join instead
                    dll_relpath = os.path.join(os.path.relpath(module_folder, os.path.dirname(csproj_file)), 'bin', file)

                    # Validate if dll already exists in the csproj
                    dll_exists = tree.find("./{val}ItemGroup/{val}Reference/{val}HintPath[.='{dll}']".format(val='{' + MS_BUILD_NAMESPACE + '}', dll=dll_relpath)) is not None

                    # Continue if dll is csproj target assembly
                    # Continue if dll already exists in csproj
                    # Continue if dll exists in blacklist
                    if dll_name == os.path.splitext(os.path.basename(csproj_file))[0] or dll_exists or dll_name in ASSEMBLY_BLACKLIST:
                        continue

                    # Create Reference structure
                    ref = ET.Element("Reference")
                    ref.set("Include", dll_name)

                    # Create Name structure
                    ref_name = ET.Element("Name")
                    ref_name.text = dll_name
                    ref.append(ref_name)

                    # Create HintPath structure
                    ref_hintpath = ET.Element("HintPath")
                    # Identify the relative path between the module's full dir and csproj file full dir
                    # Adds to the element text the dll relative path
                    ref_hintpath.text = dll_relpath
                    ref.append(ref_hintpath)

                    # Create Private structure
                    ref_private = ET.Element("Private")
                    ref_private.text = "False"
                    ref.append(ref_private)

                    # Append new element to ItemGroup element
                    itemgroup_elem.append(ref)

            # Save newly added references to csproj file
            tree.write(csproj_file)

        # Check if csproj is for the module's main assembly
        elif csproj["IsMain"] and remove_resources_files:
            # Read csproj file and find all ItemGroup elements
            ET.register_namespace('', MS_BUILD_NAMESPACE)
            tree = ET.parse(csproj_file)
            itemgroup_elems = tree.findall("./{val}ItemGroup".format(val='{' + MS_BUILD_NAMESPACE + '}'))

            # Iterate through all ItemGroups to find out which has embedded resource files
            for itemgroup in itemgroup_elems:
                emb_resource_elems = itemgroup.findall("./{val}EmbeddedResource".format(val='{' + MS_BUILD_NAMESPACE + '}'))
                if emb_resource_elems:
                    for emb_resource in emb_resource_elems:
                        # Remove every embedded resource element found
                        itemgroup.remove(emb_resource)
                    break
            # Save changes to csproj file
            tree.write(csproj_file)


def main(artifact_dir: str, lt_http_proto: str, lt_url: str, lt_api_endpoint: str, lt_api_version: int, lt_token: str, target_env: str, apps: list, trigger_manifest: dict, include_test_apps: bool, friendly_package_names: bool, include_all_refs: bool, remove_resources_files: bool):

    # Builds the LifeTime endpoint
    lt_endpoint = build_lt_endpoint(lt_http_proto, lt_url, lt_api_endpoint, lt_api_version)

    # List of application names to fetch the source code from target environment
    app_list = []

    # Extract names from manifest file (when available)
    if trigger_manifest:
        for app in trigger_manifest[MANIFEST_APPLICATION_VERSIONS]:
            if include_test_apps or not app[MANIFEST_FLAG_IS_TEST_APPLICATION]:
                app_list.append(app[MANIFEST_APPLICATION_NAME])
    else:
        app_list = apps

    for app_name in app_list:
        # Request source code package creation
        pkg_details = get_environment_app_source_code(artifact_dir, lt_endpoint, lt_token, env_name=target_env, app_name=app_name)
        pkg_key = pkg_details["PackageKey"]
        print("Source code package {} started being created for application {} deployed in {} environment.".format(pkg_key, app_name, target_env), flush=True)

        # Wait for package creation to finish
        wait_counter = 0
        link_available = False
        while wait_counter < get_configuration_value("SOURCECODE_TIMEOUT_IN_SECS", SOURCECODE_TIMEOUT_IN_SECS):
            # Check current package status
            pkg_status = get_environment_app_source_code_status(artifact_dir, lt_endpoint, lt_token,
                                                                env_name=target_env, app_name=app_name, pkg_key=pkg_key)
            if pkg_status["Status"] == SOURCECODE_FINISHED_STATUS:
                # Package was created successfully
                link_available = True
                break
            elif pkg_status["Status"] == SOURCECODE_ONGOING_STATUS:
                # Package is still being created. Go back to sleep.
                sleep_value = get_configuration_value("SOURCECODE_SLEEP_PERIOD_IN_SECS", SOURCECODE_SLEEP_PERIOD_IN_SECS)
                sleep(sleep_value)
                wait_counter += sleep_value
                print("{} secs have passed while source code package is being created...".format(wait_counter), flush=True)
            else:
                raise NotImplementedError("Unknown source code package status: {}.".format(pkg_status["Status"]))

        # When the package is created, download it using the provided key
        if link_available:
            print("Source code package {} created successfully.".format(pkg_key), flush=True)
            pkg_link = get_environment_app_source_code_link(artifact_dir, lt_endpoint, lt_token,
                                                            env_name=target_env, app_name=app_name, pkg_key=pkg_key)
            if friendly_package_names:
                target_env_key = get_environment_key(artifact_dir, lt_endpoint, lt_token, target_env)
                running_version = get_running_app_version(artifact_dir, lt_endpoint, lt_token, target_env_key, app_name=app_name)
                file_name = "{}_v{}".format(app_name.replace(" ", "_"), running_version["Version"].replace(".", "_"))
                if running_version["IsModified"]:
                    file_name += "+"
            else:
                file_name = pkg_key
            file_name += ENVIRONMENT_SOURCECODE_DOWNLOAD_FILE
            file_path = os.path.join(artifact_dir, ENVIRONMENT_SOURCECODE_FOLDER, file_name)

            download_package(file_path, lt_token, pkg_link["url"])
            print("Source code package {} downloaded successfully.".format(pkg_key), flush=True)

            # Extract source code for each module from downloaded package, applying post-processing actions (if requested)
            module_count = extract_package_content(file_path, include_all_refs, remove_resources_files)
            print("{} application modules processed successfully.".format(module_count), flush=True)
        else:
            print("Timeout expired while generating source code package {}. Unable to download source code for application {}.".format(pkg_key, app_name), flush=True)


# End of main()


if __name__ == "__main__":
    # Argument menu / parsing
    parser = argparse.ArgumentParser()
    parser.add_argument("-a", "--artifacts", type=str, default=ARTIFACT_FOLDER,
                        help="Name of the artifacts folder. Default: \"Artifacts\"")
    parser.add_argument("-lu", "--lt_url", type=str, required=True,
                        help="URL for LifeTime environment, without the API endpoint. Example: \"https://<lifetime_host>\"")
    parser.add_argument("-lt", "--lt_token", type=str, required=True,
                        help="Token for LifeTime API calls.")
    parser.add_argument("-e", "--lt_endpoint", type=str, default=LIFETIME_API_ENDPOINT,
                        help="(optional) Used to set the API endpoint for LifeTime, without the version. Default: \"lifetimeapi/rest\"")
    parser.add_argument("-t", "--target_env", type=str, required=True,
                        help="Name, as displayed in LifeTime, of the target environment where to fetch the source code from.")
    parser.add_argument("-l", "--app_list", type=str,
                        help="Comma separated list of apps you want to fetch. Example: \"App1,App2 With Spaces,App3_With_Underscores\"")
    parser.add_argument("-f", "--manifest_file", type=str,
                        help="Manifest file (with JSON format). Contains required data used throughout the pipeline execution.")
    parser.add_argument("-i", "--include_test_apps", action='store_true',
                        help="Flag that indicates if applications marked as \"Test Application\" in the manifest are fetched as well.")
    parser.add_argument("-n", "--friendly_package_names", action='store_true',
                        help="Flag that indicates if downloaded source code packages should have a user-friendly name. Example: \"<AppName>_v1_2_1\"")
    parser.add_argument("-ref", "--include_all_refs", action='store_true',
                        help="Flag that indicates if all assemblies in the \"bin\" folder should be added as references in the .csproj file.")
    parser.add_argument("-res", "--remove_resources_files", action='store_true',
                        help="Flag that indicates if embedded resources files should be removed from the .csproj file.")
    parser.add_argument("-cf", "--config_file", type=str,
                        help="Config file path. Contains configuration values to override the default ones.")

    args = parser.parse_args()

    # Load config file if exists
    if args.config_file:
        load_configuration_file(args.config_file)
    # Parse the artifact directory
    artifact_dir = args.artifacts
    # Parse the API endpoint
    lt_api_endpoint = args.lt_endpoint
    # Parse the LT Url and split the LT hostname from the HTTP protocol
    # Assumes the default HTTP protocol = https
    lt_http_proto = LIFETIME_HTTP_PROTO
    lt_url = args.lt_url
    if lt_url.startswith("http://"):
        lt_http_proto = "http"
        lt_url = lt_url.replace("http://", "")
    else:
        lt_url = lt_url.replace("https://", "")
    if lt_url.endswith("/"):
        lt_url = lt_url[:-1]
    # Parte LT API Version
    lt_version = LIFETIME_API_VERSION
    # Parse the LT Token
    lt_token = args.lt_token
    # Parse Target Environment
    target_env = args.target_env
    # Check if either an app list or a manifest file is being provided
    if not args.app_list and not args.manifest_file:
        parser.error("either --app_list or --manifest_file must be provided as arguments")
    # Use Trigger Manifest (if available)
    if args.manifest_file:
        # Parse Trigger Manifest artifact
        trigger_manifest = load_data("", args.manifest_file)
        apps = None
    else:
        trigger_manifest = None
        # Parse App list
        _apps = args.app_list
        apps = _apps.split(',')
    # Parse Include Test Apps flag
    include_test_apps = args.include_test_apps
    # Parse Friendly Package Names flag
    friendly_package_names = args.friendly_package_names
    # Parse Include All References flag
    include_all_refs = args.include_all_refs
    # Parse Remove Resources Files flag
    remove_resources_files = args.remove_resources_files

    # Calls the main script
    main(artifact_dir, lt_http_proto, lt_url, lt_api_endpoint, lt_version, lt_token, target_env, apps, trigger_manifest, include_test_apps, friendly_package_names, include_all_refs, remove_resources_files)  # type: ignore



# Python Modules
import os

# Custom Modules
# Exceptions
from outsystems.exceptions.not_enough_permissions import NotEnoughPermissionsError
# Functions
from outsystems.architecture_dashboard.ad_base import send_get_request, build_ad_endpoint
from outsystems.file_helpers.file import store_data

# Variables
from outsystems.vars.ad_vars import AD_API_ENDPOINT, AD_API_SUCCESS_CODE, AD_HTTP_PROTO, AD_API_VERSION, \
    AD_API_UNAUTHORIZED_CODE, AD_APP_ENDPOINT, AD_APP_LIMIT_DEFAULT, AD_APP_SUCCESS_CODE, \
    AD_LEVELS_ENDPOINT, AD_LEVELS_SUCCESS_CODE, AD_CATEGORIES_ENDPOINT, AD_CATEGORIES_SUCCESS_CODE, \
    AD_API_NOT_FOUND_CODE
from outsystems.vars.file_vars import AD_FOLDER, AD_FILE_PREFIX, AD_INFRA_FILE, AD_APP_FILE, \
    AD_LEVELS_FILE, AD_CATEGORIES_FILE


# Returns the infrastructure technical debt summary
def get_infra_techdebt(artifact_dir: str, ad_api_host: str, activation_code: str, api_key: str):

    # Format the request URL to include the api endpoint
    base_url = build_ad_endpoint(AD_HTTP_PROTO, ad_api_host, AD_API_ENDPOINT, AD_API_VERSION)
    request_string = "{}/{}".format(base_url, AD_APP_ENDPOINT)
    params = {"Limit": AD_APP_LIMIT_DEFAULT}

    # Sends the request
    response = send_get_request(request_string, activation_code, api_key, params)
    status_code = int(response["http_status"])

    # Process the response based on the status code returned from the server
    if status_code == AD_API_SUCCESS_CODE:
        # Stores the result
        filename = "{}{}".format(AD_FILE_PREFIX, AD_INFRA_FILE)
        filename = os.path.join(AD_FOLDER, filename)
        store_data(artifact_dir, filename, response["response"])
        return response["response"]
    elif status_code == AD_API_UNAUTHORIZED_CODE:
        raise NotEnoughPermissionsError(
            "You don't have enough permissions to get Tecnical Debt information. Details {}".format(response["response"]))
    else:
        raise NotImplementedError(
            "There was an error. Response from server: {}".format(response))


# Returns the application technical debt summary
def get_app_techdebt(artifact_dir: str, ad_api_host: str, activation_code: str, api_key: str, app: dict):

    # Format the request URL to include the api endpoint
    base_url = build_ad_endpoint(AD_HTTP_PROTO, ad_api_host, AD_API_ENDPOINT, AD_API_VERSION)
    request_string = "{}/{}".format(base_url, AD_APP_ENDPOINT)
    params = {"ApplicationGUID": app["ApplicationKey"]}

    # Sends the request
    response = send_get_request(request_string, activation_code, api_key, params)
    status_code = int(response["http_status"])

    # Process the response based on the status code returned from the server
    if status_code == AD_APP_SUCCESS_CODE:
        # Stores the result
        filename = "{}.{}{}".format(AD_FILE_PREFIX, app["ApplicationName"], AD_APP_FILE)
        filename = os.path.join(AD_FOLDER, filename)
        store_data(artifact_dir, filename, response["response"])
        return response["response"]
    # No application found with a key matching the Application input parameter
    # Probably all modules of the app are ignored"
    elif status_code == AD_API_NOT_FOUND_CODE:
        return None
    elif status_code == AD_API_UNAUTHORIZED_CODE:
        raise NotEnoughPermissionsError(
            "You don't have enough permissions to get Tecnical Debt information. Details {}".format(response["response"]))
    else:
        raise NotImplementedError(
            "There was an error. Response from server: {}".format(response))


# Returns the technical debt levels detail
def get_techdebt_levels(artifact_dir: str, ad_api_host: str, activation_code: str, api_key: str):

    # Format the request URL to include the api endpoint
    base_url = build_ad_endpoint(AD_HTTP_PROTO, ad_api_host, AD_API_ENDPOINT, AD_API_VERSION)
    request_string = "{}/{}".format(base_url, AD_LEVELS_ENDPOINT)

    # Sends the request
    response = send_get_request(request_string, activation_code, api_key)
    status_code = int(response["http_status"])

    # Process the response based on the status code returned from the server
    if status_code == AD_LEVELS_SUCCESS_CODE:
        # Stores the result
        filename = "{}{}".format(AD_FILE_PREFIX, AD_LEVELS_FILE)
        filename = os.path.join(AD_FOLDER, filename)
        store_data(artifact_dir, filename, response["response"])
        return response["response"]
    elif status_code == AD_API_UNAUTHORIZED_CODE:
        raise NotEnoughPermissionsError(
            "You don't have enough permissions to get Tecnical Debt information. Details {}".format(response["response"]))
    else:
        raise NotImplementedError(
            "There was an error. Response from server: {}".format(response))


# Returns the technical debt categories detail
def get_techdebt_categories(artifact_dir: str, ad_api_host: str, activation_code: str, api_key: str):

    # Format the request URL to include the api endpoint
    base_url = build_ad_endpoint(AD_HTTP_PROTO, ad_api_host, AD_API_ENDPOINT, AD_API_VERSION)
    request_string = "{}/{}".format(base_url, AD_CATEGORIES_ENDPOINT)

    # Sends the request
    response = send_get_request(request_string, activation_code, api_key)
    status_code = int(response["http_status"])

    # Process the response based on the status code returned from the server
    if status_code == AD_CATEGORIES_SUCCESS_CODE:
        # Stores the result
        filename = "{}{}".format(AD_FILE_PREFIX, AD_CATEGORIES_FILE)
        filename = os.path.join(AD_FOLDER, filename)
        store_data(artifact_dir, filename, response["response"])
        return response["response"]
    elif status_code == AD_API_UNAUTHORIZED_CODE:
        raise NotEnoughPermissionsError(
            "You don't have enough permissions to get Tecnical Debt information. Details {}".format(response["response"]))
    else:
        raise NotImplementedError(
            "There was an error. Response from server: {}".format(response))



# Python Modules
import os

# Custom Modules
# Exceptions
from outsystems.exceptions.invalid_parameters import InvalidParametersError
from outsystems.exceptions.environment_not_found import EnvironmentNotFoundError
from outsystems.exceptions.not_enough_permissions import NotEnoughPermissionsError
from outsystems.exceptions.server_error import ServerError
# Functions
from outsystems.lifetime.lifetime_base import send_download_request

# Variables
from outsystems.vars.lifetime_vars import DOWNLOAD_SUCCESS_CODE, DOWNLOAD_INVALID_KEY_CODE, \
    DOWNLOAD_NO_PERMISSION_CODE, DOWNLOAD_NOT_FOUND, DOWNLOAD_FAILED_CODE


# Downloads a binary file from a LifeTime download link
def download_package(file_path: str, auth_token: str, pkg_url: str):
    # Sends the request
    response = send_download_request(pkg_url, auth_token)
    status_code = int(response["http_status"])

    if status_code == DOWNLOAD_SUCCESS_CODE:
        # Remove the spaces in the filename
        file_path = file_path.replace(" ", "_")
        # Makes sure that, if a directory is in the filename, that directory exists
        os.makedirs(os.path.dirname(file_path), exist_ok=True)
        with open(file_path, "wb") as f:
            f.write(response["response"])
    elif status_code == DOWNLOAD_INVALID_KEY_CODE:
        raise InvalidParametersError("The required type <Type> is invalid for given keys (EnvironmentKey; ApplicationKey). Details: {}".format(
            response["response"]))
    elif status_code == DOWNLOAD_NO_PERMISSION_CODE:
        raise NotEnoughPermissionsError("User doesn't have permissions for the given keys (EnvironmentKey; ApplicationKey). Details: {}".format(
            response["response"]))
    elif status_code == DOWNLOAD_NOT_FOUND:
        raise EnvironmentNotFoundError("No environment or application found. Please check that the EnvironmentKey and ApplicationKey exist. Details: {}".format(
            response["response"]))
    elif status_code == DOWNLOAD_FAILED_CODE:
        raise ServerError("Failed to start the operation to package. Details: {}".format(
            response["response"]))
    else:
        raise NotImplementedError(
            "There was an error. Response from server: {}".format(response))



# Custom Modules
# Functions
from outsystems.bdd_framework.bdd_base import send_bdd_get_run_request
# Variables
from outsystems.vars.bdd_vars import BDD_RUNNER_SUCCESS_CODE


# Run existing BDD test in the target environment.
def run_bdd_test(test_url: str):
    # Sends the request
    response = send_bdd_get_run_request(test_url, None)
    status_code = response["http_status"]
    if status_code == BDD_RUNNER_SUCCESS_CODE:
        return response["response"]
    else:
        raise NotImplementedError(
            "There was an error. Response from server: {}".format(response))



# Python Modules
import sys
import os
import argparse
from time import sleep

# Workaround for Jenkins:
# Set the path to include the outsystems module
# Jenkins exposes the workspace directory through env.
if "WORKSPACE" in os.environ:
    sys.path.append(os.environ['WORKSPACE'])
else:  # Else just add the project dir
    sys.path.append(os.getcwd())

# Custom Modules
# Variables
from outsystems.vars.file_vars import ARTIFACT_FOLDER
from outsystems.vars.lifetime_vars import LIFETIME_HTTP_PROTO, LIFETIME_API_ENDPOINT, LIFETIME_API_VERSION
from outsystems.vars.pipeline_vars import QUEUE_TIMEOUT_IN_SECS, SLEEP_PERIOD_IN_SECS, CONFLICTS_FILE, \
    REDEPLOY_OUTDATED_APPS, DEPLOYMENT_TIMEOUT_IN_SECS, DEPLOYMENT_RUNNING_STATUS, DEPLOYMENT_WAITING_STATUS, \
    DEPLOYMENT_ERROR_STATUS_LIST, DEPLOY_ERROR_FILE, ALLOW_CONTINUE_WITH_ERRORS
# Functions
from outsystems.lifetime.lifetime_environments import get_environment_key
from outsystems.lifetime.lifetime_deployments import get_deployment_status, get_deployment_info, \
    send_binary_deployment, delete_deployment, start_deployment, continue_deployment, get_running_deployment, \
    check_deployment_two_step_deploy_status
from outsystems.file_helpers.file import store_data, is_valid_os_package
from outsystems.lifetime.lifetime_base import build_lt_endpoint
from outsystems.vars.vars_base import get_configuration_value, load_configuration_file
# Exceptions
from outsystems.exceptions.invalid_os_package import InvalidOutSystemsPackage


# ############################################################# SCRIPT ##############################################################
def main(artifact_dir: str, lt_http_proto: str, lt_url: str, lt_api_endpoint: str, lt_api_version: int, lt_token: str, dest_env_label: str, force_two_step_deployment: bool, package_path: str, allow_parallel_deployments: bool):

    # Builds the LifeTime endpoint
    lt_endpoint = build_lt_endpoint(lt_http_proto, lt_url, lt_api_endpoint, lt_api_version)

    # Gets the environment key for the destination environment
    dest_env_key = get_environment_key(artifact_dir, lt_endpoint, lt_token, dest_env_label)

    if not allow_parallel_deployments:
        wait_counter = 0
        deployments = get_running_deployment(artifact_dir, lt_endpoint, lt_token, dest_env_key)
        while len(deployments) > 0:
            if wait_counter >= get_configuration_value("QUEUE_TIMEOUT_IN_SECS", QUEUE_TIMEOUT_IN_SECS):
                print("Timeout occurred while waiting for LifeTime to be free, to create the new deployment plan.", flush=True)
                sys.exit(1)
            sleep_value = get_configuration_value("SLEEP_PERIOD_IN_SECS", SLEEP_PERIOD_IN_SECS)
            sleep(sleep_value)
            wait_counter += sleep_value
            print("Waiting for LifeTime to be free. Elapsed time: {} seconds...".format(wait_counter), flush=True)
            deployments = get_running_deployment(artifact_dir, lt_endpoint, lt_token, dest_env_key)

    # LT is free to deploy
    # Validate if file has OutSystems package extension
    if not is_valid_os_package(package_path):
        raise InvalidOutSystemsPackage("Binary file is not an OutSystems package. Expected 'osp' or 'oap' as file extension.")

    # Send the deployment plan and grab the key
    dep_plan_key = send_binary_deployment(artifact_dir, lt_endpoint, lt_token, lt_api_version, dest_env_key, package_path)
    print("Deployment plan {} created successfully.".format(dep_plan_key), flush=True)

    # Check if created deployment plan has conflicts
    dep_details = get_deployment_info(artifact_dir, lt_endpoint, lt_token, dep_plan_key)
    has_conflicts = len(dep_details["ApplicationConflicts"]) > 0
    if has_conflicts:
        store_data(artifact_dir, CONFLICTS_FILE, dep_details["ApplicationConflicts"])
        if not get_configuration_value("ALLOW_CONTINUE_WITH_ERRORS", ALLOW_CONTINUE_WITH_ERRORS) or lt_api_version == 1:
            print("Deployment plan {} has conflicts and will be aborted. Check {} artifact for more details.".format(dep_plan_key, CONFLICTS_FILE), flush=True)
            # Abort previously created deployment plan to target environment
            delete_deployment(lt_endpoint, lt_token, dep_plan_key)
            print("Deployment plan {} was deleted successfully.".format(dep_plan_key), flush=True)
            sys.exit(1)
        else:
            print("Deployment plan {} has conflicts but will continue with errors. Check {} artifact for more details.".format(dep_plan_key, CONFLICTS_FILE), flush=True)

    # Check if outdated consumer applications (outside of deployment plan) should be redeployed and start the deployment plan execution
    if lt_api_version == 1:  # LT for OS version < 11
        start_deployment(lt_endpoint, lt_token, dep_plan_key)
    elif lt_api_version == 2:  # LT for OS v11
        if has_conflicts:
            start_deployment(lt_endpoint, lt_token, dep_plan_key, redeploy_outdated=False, continue_with_errors=True)
        else:
            start_deployment(lt_endpoint, lt_token, dep_plan_key, redeploy_outdated=get_configuration_value("REDEPLOY_OUTDATED_APPS", REDEPLOY_OUTDATED_APPS))
    else:
        raise NotImplementedError("Please make sure the API version is compatible with the module.")
    print("Deployment plan {} started being executed.".format(dep_plan_key), flush=True)

    # Flag to only alert the user once
    alert_user = False
    # Sleep thread until deployment has finished
    wait_counter = 0
    while wait_counter < get_configuration_value("DEPLOYMENT_TIMEOUT_IN_SECS", DEPLOYMENT_TIMEOUT_IN_SECS):
        # Check Deployment Plan status.
        dep_status = get_deployment_status(
            artifact_dir, lt_endpoint, lt_token, dep_plan_key)
        if dep_status["DeploymentStatus"] != DEPLOYMENT_RUNNING_STATUS:
            # Check deployment status is pending approval.
            if dep_status["DeploymentStatus"] == DEPLOYMENT_WAITING_STATUS:
                # Check if deployment waiting status is due to 2-Step
                if check_deployment_two_step_deploy_status(dep_status):
                    # Force it to continue in case of force_two_step_deployment parameter
                    if force_two_step_deployment:
                        continue_deployment(lt_endpoint, lt_token, dep_plan_key)
                        print("Deployment plan {} resumed execution.".format(dep_plan_key), flush=True)
                    else:
                        # Exit the script to continue with the pipeline execution
                        print("Deployment plan {} first step finished successfully.".format(dep_plan_key), flush=True)
                        sys.exit(0)
                # Send notification to alert deployment manual intervention.
                elif not alert_user:
                    alert_user = True
                    print("A manual intervention is required to continue the execution of the deployment plan {}.".format(dep_plan_key), flush=True)
            elif dep_status["DeploymentStatus"] in DEPLOYMENT_ERROR_STATUS_LIST:
                print("Deployment plan finished with status {}.".format(dep_status["DeploymentStatus"]), flush=True)
                store_data(artifact_dir, DEPLOY_ERROR_FILE, dep_status)
                sys.exit(1)
            else:
                # If it reaches here, it means the deployment was successful
                print("Deployment plan finished with status {}.".format(dep_status["DeploymentStatus"]), flush=True)
                # Exit the script to continue with the pipeline
                sys.exit(0)
        # Deployment status is still running. Go back to sleep.
        sleep_value = get_configuration_value("SLEEP_PERIOD_IN_SECS", SLEEP_PERIOD_IN_SECS)
        sleep(sleep_value)
        wait_counter += sleep_value
        print("{} secs have passed since the deployment started...".format(wait_counter), flush=True)

    # Deployment timeout reached. Exit script with error
    print("Timeout occurred while deployment plan is still in {} status.".format(DEPLOYMENT_RUNNING_STATUS), flush=True)
    sys.exit(1)


# End of main()


if __name__ == "__main__":
    # Argument menu / parsing
    parser = argparse.ArgumentParser()
    parser.add_argument("-a", "--artifacts", type=str, default=ARTIFACT_FOLDER,
                        help="(Optional) Name of the artifacts folder. Default: \"Artifacts\"")
    parser.add_argument("-u", "--lt_url", type=str, required=True,
                        help="URL for LifeTime environment, without the API endpoint. Example: \"https://<lifetime_host>\"")
    parser.add_argument("-t", "--lt_token", type=str, required=True,
                        help="Token for LifeTime API calls.")
    parser.add_argument("-v", "--lt_api_version", type=int, default=LIFETIME_API_VERSION,
                        help="(Optional) LifeTime API version number. If version <= 10, use 1, if version >= 11, use 2. Default: 2")
    parser.add_argument("-e", "--lt_endpoint", type=str, default=LIFETIME_API_ENDPOINT,
                        help="(Optional) Used to set the API endpoint for LifeTime, without the version. Default: \"lifetimeapi/rest\"")
    parser.add_argument("-d", "--destination_env_label", type=str, required=True,
                        help="Label, as configured in the manifest, of the destination environment where you want to deploy the apps.")
    parser.add_argument("-p", "--package_path", type=str, required=True,
                        help="Package file path")
    parser.add_argument("-c", "--force_two_step_deployment", action='store_true',
                        help="Force the execution of the 2-Step deployment.")
    parser.add_argument("-cf", "--config_file", type=str,
                        help="Config file path. Contains configuration values to override the default ones.")
    parser.add_argument("-pd", "--allow_parallel_deployments", action='store_true',
                        help="Skip LifeTime validation for active deployment plans.")

    args = parser.parse_args()

    # Load config file if exists
    if args.config_file:
        load_configuration_file(args.config_file)
    # Parse the artifact directory
    artifact_dir = args.artifacts
    # Parse the API endpoint
    lt_api_endpoint = args.lt_endpoint
    # Parse the LT Url and split the LT hostname from the HTTP protocol
    # Assumes the default HTTP protocol = https
    lt_http_proto = LIFETIME_HTTP_PROTO
    lt_url = args.lt_url
    if lt_url.startswith("http://"):
        lt_http_proto = "http"
        lt_url = lt_url.replace("http://", "")
    else:
        lt_url = lt_url.replace("https://", "")
    if lt_url.endswith("/"):
        lt_url = lt_url[:-1]
    # Parte LT API Version
    lt_version = args.lt_api_version
    # Parse the LT Token
    lt_token = args.lt_token
    # Parse Destination Environment
    dest_env_label = args.destination_env_label
    # Parse the package directory
    package_path = args.package_path
    # Parse Force Two-step Deployment flag
    force_two_step_deployment = args.force_two_step_deployment
    # Parse Allow Parallel Deployments
    allow_parallel_deployments = args.allow_parallel_deployments

    # Calls the main script
    main(artifact_dir, lt_http_proto, lt_url, lt_api_endpoint, lt_version, lt_token, dest_env_label, force_two_step_deployment, package_path, allow_parallel_deployments)



# Python Modules
import sys
import os
import argparse

# Workaround for Jenkins:
# Set the path to include the outsystems module
# Jenkins exposes the workspace directory through env.
if "WORKSPACE" in os.environ:
    sys.path.append(os.environ['WORKSPACE'])
else:  # Else just add the project dir
    sys.path.append(os.getcwd())

# Custom Modules
# Variables
from outsystems.vars.file_vars import ARTIFACT_FOLDER
from outsystems.vars.manifest_vars import MANIFEST_APPLICATION_VERSIONS
from outsystems.vars.ad_vars import AD_API_HOST

# Functions
from outsystems.file_helpers.file import load_data
from outsystems.architecture_dashboard.ad_tech_debt import get_infra_techdebt, get_app_techdebt, \
    get_techdebt_levels, get_techdebt_categories


# ############################################################# SCRIPT ##############################################################
def main(artifact_dir: str, ad_api_host: str, activation_code: str, api_key: str, trigger_manifest: dict):

    # Get tech debt reference data (levels)
    get_techdebt_levels(artifact_dir, ad_api_host, activation_code, api_key)
    print("Technical debt levels retrieved successfully.", flush=True)

    # Get tech debt reference data (categories)
    get_techdebt_categories(artifact_dir, ad_api_host, activation_code, api_key)
    print("Technical debt categories retrieved successfully.", flush=True)

    # If the manifest file is being used, tech debt analysis is made for each app in the manifest
    # Otherwise it runs for the entire infrastructure
    if trigger_manifest and MANIFEST_APPLICATION_VERSIONS in trigger_manifest:
        for app in trigger_manifest[MANIFEST_APPLICATION_VERSIONS]:
            status = get_app_techdebt(artifact_dir, ad_api_host, activation_code, api_key, app)
            if status:
                print("Technical debt data retrieved successfully for application {}.".format(app["ApplicationName"]), flush=True)
            else:
                print("No technical debt data found for application {}.".format(app["ApplicationName"]), flush=True)

    else:
        get_infra_techdebt(artifact_dir, ad_api_host, activation_code, api_key)
        print("Technical debt data retrieved successfully for infrastructure {}.".format(activation_code), flush=True)

    sys.exit(0)

# End of main()


if __name__ == "__main__":
    # Argument menu / parsing
    parser = argparse.ArgumentParser()
    parser.add_argument("-a", "--artifacts", type=str, default=ARTIFACT_FOLDER,
                        help="Name of the artifacts folder. Default: \"Artifacts\"")
    parser.add_argument("-n", "--ad_hostname", type=str, default=AD_API_HOST,
                        help="Hostname of Architecture Dashboard, without the API endpoint. Default: \"architecture.outsystems.com\"")
    parser.add_argument("-c", "--activation_code", type=str, required=True,
                        help="Activation code of target infrastructure.")
    parser.add_argument("-k", "--api_key", type=str, required=True,
                        help="Key for Architecture Dashboard API calls.")
    parser.add_argument("-f", "--manifest_file", type=str,
                        help="(Optional) Trigger manifest file path.")

    args = parser.parse_args()

    # Parse the artifact directory
    artifact_dir = args.artifacts
    # Parse the Architecture Dashboard hostname
    ad_api_host = args.ad_hostname
    # Parse the Infrastcucture Activation Code
    activation_code = args.activation_code
    # Parse the Architecture Dashboard API Key
    api_key = args.api_key
    # Parse Manifest file if it exists
    if args.manifest_file:
        manifest_path = os.path.split(args.manifest_file)
        manifest_file = load_data(manifest_path[0], manifest_path[1])
    else:
        manifest_file = None

    # Calls the main script
    main(artifact_dir, ad_api_host, activation_code, api_key, manifest_file)



# Python Modules
import sys
import os
import argparse
import json

# Workaround for Jenkins:
# Set the path to include the outsystems module
# Jenkins exposes the workspace directory through env.
if "WORKSPACE" in os.environ:
    sys.path.append(os.environ['WORKSPACE'])
else:  # Else just add the project dir
    sys.path.append(os.getcwd())

# Custom Modules
# Variables
from outsystems.vars.file_vars import ARTIFACT_FOLDER
from outsystems.vars.lifetime_vars import LIFETIME_HTTP_PROTO, LIFETIME_API_ENDPOINT, LIFETIME_API_VERSION
from outsystems.vars.manifest_vars import MANIFEST_APPLICATION_VERSIONS, MANIFEST_FLAG_IS_TEST_APPLICATION, MANIFEST_APPLICATION_NAME
# Functions
from outsystems.lifetime.lifetime_applications import get_applications
from outsystems.lifetime.lifetime_environments import get_environment_key
from outsystems.file_helpers.file import load_data
from outsystems.lifetime.lifetime_base import build_lt_endpoint
from outsystems.vars.vars_base import load_configuration_file
# Exceptions
from outsystems.exceptions.app_does_not_exist import AppDoesNotExistError
from outsystems.exceptions.manifest_does_not_exist import ManifestDoesNotExistError


def main(artifact_dir: str, lt_http_proto: str, lt_url: str, lt_api_endpoint: str, lt_api_version: int, lt_token: str, env_label: str, include_test_apps: bool, trigger_manifest: dict):

    # Builds the LifeTime endpoint
    lt_endpoint = build_lt_endpoint(lt_http_proto, lt_url, lt_api_endpoint, lt_api_version)

    # Gets the environment key for the target environment
    env_key = get_environment_key(artifact_dir, lt_endpoint, lt_token, env_label)

    # Get Applications without extra data
    apps = get_applications(artifact_dir, lt_endpoint, lt_token, True)
    print("OS Applications data retrieved successfully.", flush=True)

    app_names_to_validate = [app[MANIFEST_APPLICATION_NAME] for app in trigger_manifest.get(MANIFEST_APPLICATION_VERSIONS, []) if include_test_apps or not app.get(MANIFEST_FLAG_IS_TEST_APPLICATION)]

    # Check if all manifest application names exist in the infra
    if not all(any(app["Name"] == app_name for app in apps) for app_name in app_names_to_validate):
        raise AppDoesNotExistError("One or more applications not found in this infra.")

    # Check if the each manifest application exists in the provided environment
    for app_info in apps:
        if app_info["Name"] in app_names_to_validate:
            app_status_in_envs = app_info.get("AppStatusInEnvs", [])
            environment_keys = [env["EnvironmentKey"] for env in app_status_in_envs]
            if env_key not in environment_keys:
                raise AppDoesNotExistError("Application '{}' does not exist in '{}' Environment .".format(app_info['Name'], env_key))

    print("All Trigger Manifest Applications exist in the {} Environment.".format(env_label), flush=True)

# End of main()


if __name__ == "__main__":
    # Argument menu / parsing
    parser = argparse.ArgumentParser()
    parser.add_argument("-a", "--artifacts", type=str, default=ARTIFACT_FOLDER,
                        help="(Optional) Name of the artifacts folder. Default: \"Artifacts\"")
    parser.add_argument("-u", "--lt_url", type=str, required=True,
                        help="URL for LifeTime environment, without the API endpoint. Example: \"https://<lifetime_host>\"")
    parser.add_argument("-t", "--lt_token", type=str, required=True,
                        help="Token for LifeTime API calls.")
    parser.add_argument("-v", "--lt_api_version", type=int, default=LIFETIME_API_VERSION,
                        help="(Optional) LifeTime API version number. If version <= 10, use 1, if version >= 11, use 2. Default: 2")
    parser.add_argument("-e", "--lt_endpoint", type=str, default=LIFETIME_API_ENDPOINT,
                        help="(Optional) Used to set the API endpoint for LifeTime, without the version. Default: \"lifetimeapi/rest\"")
    parser.add_argument("-s", "--env_label", type=str, required=True,
                        help="Label, as configured in the manifest, of the source environment where the apps are.")
    parser.add_argument("-i", "--include_test_apps", action='store_true',
                        help="Flag that indicates if applications marked as \"Test Application\" in the manifest are included in the deployment plan.")
    parser.add_argument("-m", "--trigger_manifest", type=str,
                        help="Manifest artifact (in JSON format) received when the pipeline is triggered. Contains required data used throughout the pipeline execution.")
    parser.add_argument("-f", "--manifest_file", type=str,
                        help="Manifest file (with JSON format). Contains required data used throughout the pipeline execution.")
    parser.add_argument("-cf", "--config_file", type=str,
                        help="Config file path. Contains configuration values to override the default ones.")

    args = parser.parse_args()

    # Load config file if exists
    if args.config_file:
        load_configuration_file(args.config_file)
    # Parse the artifact directory
    artifact_dir = args.artifacts
    # Parse the API endpoint
    lt_api_endpoint = args.lt_endpoint
    # Parse the LT Url and split the LT hostname from the HTTP protocol
    # Assumes the default HTTP protocol = https
    lt_http_proto = LIFETIME_HTTP_PROTO
    lt_url = args.lt_url
    if lt_url.startswith("http://"):
        lt_http_proto = "http"
        lt_url = lt_url.replace("http://", "")
    else:
        lt_url = lt_url.replace("https://", "")
    if lt_url.endswith("/"):
        lt_url = lt_url[:-1]
    # Parte LT API Version
    lt_version = args.lt_api_version
    # Parse the LT Token
    lt_token = args.lt_token
    # Parse Environment
    env_label = args.env_label
    # Parse Include Test Apps flag
    include_test_apps = args.include_test_apps

    # Validate Manifest is being passed either as JSON or as file
    if not args.trigger_manifest and not args.manifest_file:
        raise ManifestDoesNotExistError("The manifest was not provided as JSON or as a file. Aborting!")

    # Parse Trigger Manifest artifact
    if args.manifest_file:
        trigger_manifest_path = os.path.split(args.manifest_file)
        trigger_manifest = load_data(trigger_manifest_path[0], trigger_manifest_path[1])
    else:
        trigger_manifest = json.loads(args.trigger_manifest)

    # Calls the main script
    main(artifact_dir, lt_http_proto, lt_url, lt_api_endpoint, lt_version, lt_token, env_label, include_test_apps, trigger_manifest)



# Python Modules
import sys
import os
import argparse

# Workaround for Jenkins:
# Set the path to include the outsystems module
# Jenkins exposes the workspace directory through env.
if "WORKSPACE" in os.environ:
    sys.path.append(os.environ['WORKSPACE'])
else:  # Else just add the project dir
    sys.path.append(os.getcwd())

# Custom Modules
# Variables
from outsystems.vars.file_vars import ARTIFACT_FOLDER, SOLUTIONS_FOLDER, SOLUTIONS_DEPLOY_FILE
# Functions
from outsystems.osp_tool.osp_base import call_osptool
from outsystems.vars.vars_base import load_configuration_file
from outsystems.file_helpers.file import store_data
# Exceptions
from outsystems.exceptions.osptool_error import OSPToolDeploymentError


# ############################################################# SCRIPT ##############################################################
def main(artifact_dir: str, dest_env: str, package_path: str, catalogmappings_path: str, osp_tool_path: str, credentials: str):

    # Get solution file name from path
    solution_file = os.path.split(package_path)[1]

    print("Starting deployment of '{}' into '{}' environment...".format(solution_file, dest_env), flush=True)

    # Call OSP Tool
    return_code, execution_log = call_osptool(osp_tool_path, package_path, dest_env, credentials, catalogmappings_path)

    # Split the output into lines
    execution_log = execution_log.splitlines()

    # Stores the execution log
    filename = "{}{}".format(solution_file, SOLUTIONS_DEPLOY_FILE)
    filename = os.path.join(SOLUTIONS_FOLDER, filename)
    store_data(artifact_dir, filename, execution_log)

    error_validation_list = ['Incompatible Dependency', 'Execution Plan Abort', 'Outdated Consumer', 'Missing Configuration']

    # Validate the presence of each error validation
    deploy_error_flag = False
    for error_validation in error_validation_list:
        existing_error_list = [s for s in execution_log if error_validation in s]
        if existing_error_list:
            deploy_error_flag = True
            print(f'\nFound "{error_validation}" validation:')
            for error in existing_error_list:
                print(f' - {error}')

    if deploy_error_flag:
        # Exit script with error
        raise OSPToolDeploymentError(
            "OSP Tool Deployment finished with errors. Please check the logs for further details.")


if __name__ == "__main__":
    # Argument menu / parsing
    parser = argparse.ArgumentParser()
    parser.add_argument("-a", "--artifacts", type=str, default=ARTIFACT_FOLDER,
                        help="Name of the artifacts folder. Default: \"Artifacts\"")
    parser.add_argument("-d", "--destination_env", type=str, required=True,
                        help="Name, as displayed in LifeTime, of the destination environment where you want to deploy the apps. (if in Airgap mode should be the hostname of the destination environment where you want to deploy the apps)")
    parser.add_argument("-p", "--package_path", type=str, required=True,
                        help="Package file path")
    parser.add_argument("-c", "--catalogmappings_path", type=str,
                        help="(Optional) Catalog mappings file path")
    parser.add_argument("-o", "--osp_tool_path", type=str, required=True,
                        help="OSP Tool file path")
    parser.add_argument("-user", "--osptool_user", type=str, required=True,
                        help="Username with privileges to deploy applications on target environment")
    parser.add_argument("-pwd", "--osptool_pwd", type=str, required=True,
                        help="Password of the Username with priveleges to deploy applications on target environment")
    parser.add_argument("-cf", "--config_file", type=str,
                        help="Config file path. Contains configuration values to override the default ones.")

    args = parser.parse_args()

    # Load config file if exists
    if args.config_file:
        load_configuration_file(args.config_file)
    # Parse the artifact directory
    artifact_dir = args.artifacts
    # Parse the package path
    package_path = args.package_path
    # Parse the Catalog Mapping path
    catalogmappings_path = args.catalogmappings_path
    # Parse Destination Environment
    dest_env = args.destination_env
    # Parse OSP Tool path
    osp_tool_path = args.osp_tool_path
    # Parse Credentials for OSP Tool
    credentials = args.osptool_user + " " + args.osptool_pwd

    # Calls the main script
    main(artifact_dir, dest_env, package_path, catalogmappings_path, osp_tool_path, credentials)



# Deployment specific variables
QUEUE_TIMEOUT_IN_SECS = 1800
DEPLOYMENT_TIMEOUT_IN_SECS = 3600
SLEEP_PERIOD_IN_SECS = 20
REDEPLOY_OUTDATED_APPS = True
ALLOW_CONTINUE_WITH_ERRORS = False
DEPLOYMENT_STATUS_LIST = ["saved", "running", "needs_user_intervention", "aborting"]
DEPLOYMENT_ERROR_STATUS_LIST = ["aborted", "finished_with_errors"]
DEPLOYMENT_WAITING_STATUS = "needs_user_intervention"
DEPLOYMENT_RUNNING_STATUS = "running"
DEPLOYMENT_SAVED_STATUS = "saved"

# Pipeline files variables
CONFLICTS_FILE = "DeploymentConflicts"
DEPLOY_ERROR_FILE = "DeploymentErrors"

# Application specific variables
MAX_VERSIONS_TO_RETURN = 10
TAG_APP_MAX_RETRIES = 5

# Environment specific variables
SOURCECODE_TIMEOUT_IN_SECS = 3600
SOURCECODE_SLEEP_PERIOD_IN_SECS = 10
SOURCECODE_ONGOING_STATUS = "InProgress"
SOURCECODE_FINISHED_STATUS = "Done"

# Solutions specific variables
SOLUTION_TIMEOUT_IN_SECS = 3600
SOLUTION_SLEEP_PERIOD_IN_SECS = 2
SOLUTION_CREATED_STATUS = "Created"
SOLUTION_READY_STATUS = "Ready"
SOLUTION_GATHERING_DEPENDENCIES_STATUS = "Gathering Dependencies"
SOLUTION_GETTING_BINARIES_STATUS = "Getting Binaries"
SOLUTION_GENERATING_META_MODEL_STATUS = "Generating Meta Model"
SOLUTION_GENERATING_SOLUTION_STATUS = "Generating Solution"
SOLUTION_COMPLETED_STATUS = "Completed"
SOLUTION_ABORTED_STATUS = "Aborted"



# Python Modules
import sys
import os
import argparse
from packaging.version import Version
from time import sleep
import json

# Workaround for Jenkins:
# Set the path to include the outsystems module
# Jenkins exposes the workspace directory through env.
if "WORKSPACE" in os.environ:
    sys.path.append(os.environ['WORKSPACE'])
else:  # Else just add the project dir
    sys.path.append(os.getcwd())

# Custom Modules
# Variables
from outsystems.vars.file_vars import ARTIFACT_FOLDER
from outsystems.vars.lifetime_vars import LIFETIME_HTTP_PROTO, LIFETIME_API_ENDPOINT, LIFETIME_API_VERSION
from outsystems.vars.manifest_vars import MANIFEST_APPLICATION_VERSIONS, MANIFEST_FLAG_IS_TEST_APPLICATION
from outsystems.vars.pipeline_vars import QUEUE_TIMEOUT_IN_SECS, SLEEP_PERIOD_IN_SECS, CONFLICTS_FILE, \
    REDEPLOY_OUTDATED_APPS, DEPLOYMENT_TIMEOUT_IN_SECS, DEPLOYMENT_RUNNING_STATUS, DEPLOYMENT_WAITING_STATUS, \
    DEPLOYMENT_ERROR_STATUS_LIST, DEPLOY_ERROR_FILE, ALLOW_CONTINUE_WITH_ERRORS
# Functions
from outsystems.lifetime.lifetime_environments import get_environment_app_version, get_environment_deployment_zones
from outsystems.lifetime.lifetime_applications import get_application_version
from outsystems.lifetime.lifetime_deployments import get_deployment_status, get_deployment_info, \
    send_deployment, delete_deployment, start_deployment, continue_deployment, get_running_deployment, \
    check_deployment_two_step_deploy_status
from outsystems.file_helpers.file import store_data, load_data
from outsystems.lifetime.lifetime_base import build_lt_endpoint
from outsystems.manifest.manifest_base import get_environment_details, get_deployment_notes
from outsystems.vars.vars_base import get_configuration_value, load_configuration_file
# Exceptions
from outsystems.exceptions.app_does_not_exist import AppDoesNotExistError
from outsystems.exceptions.manifest_does_not_exist import ManifestDoesNotExistError


# ############################################################# SCRIPT ##############################################################
# Function that will generate the app key portion of the deployment for LifeTime, based on the API level
def generate_deploy_app_key(lt_api_version: int, app_version_key: str, deploy_zone=""):
    if lt_api_version == 1:  # LT for OS version < 11
        return app_version_key
    elif lt_api_version == 2:  # LT for OS v11
        return {"ApplicationVersionKey": app_version_key, "DeploymentZoneKey": deploy_zone}
    else:
        raise NotImplementedError("Please make sure the API version is compatible with the module.")


# Function that will build the info required for a deployment based on a manifest file
def generate_deployment_based_on_manifest(artifact_dir: str, lt_endpoint: str, lt_token: str, src_env_key: str, src_env_name: str, manifest: list, include_test_apps: bool, include_deployment_zones: bool):
    app_data_list = []  # will contain the applications details from the manifest

    for deployed_app in manifest[MANIFEST_APPLICATION_VERSIONS]:  # type: ignore
        if not include_test_apps and deployed_app[MANIFEST_FLAG_IS_TEST_APPLICATION]:
            continue
        try:
            get_application_version(artifact_dir, lt_endpoint, lt_token, False, deployed_app["VersionKey"], app_name=deployed_app["ApplicationName"])
        except AppDoesNotExistError:
            print("Application {} with version {} no longer exists in {}. The manifest no longer reflects the current state of the environment. Aborting!".format(deployed_app["ApplicationName"], deployed_app["VersionNumber"], src_env_name), flush=True)
            sys.exit(1)
        except Exception as error:
            print("Error trying to validate if the application {} exists in the {} environment.\nError: {}".format(deployed_app["ApplicationName"], src_env_name, error), flush=True)
            sys.exit(1)

        # Add it to the app data list
        app_data = {'Name': deployed_app["ApplicationName"], 'Key': deployed_app["ApplicationKey"], 'Version': deployed_app["VersionNumber"], 'VersionKey': deployed_app["VersionKey"]}
        if include_deployment_zones:
            app_data['DeploymentZone'] = deployed_app["DeploymentZoneName"]
        app_data_list.append(app_data)

    return app_data_list


# Function to check if target environment already has the application versions to be deployed
def check_if_can_deploy(artifact_dir: str, lt_endpoint: str, lt_api_version: str, lt_token: str, env_key: str, env_name: str, app_data_list: list, include_deployment_zones: bool):
    app_keys = []  # will contain the application keys to create the deployment plan
    deploy_zones = []  # will contain the deployment zones available in the target environment

    # Get information about the deployment zones in the target environment
    if include_deployment_zones:
        deploy_zones = get_environment_deployment_zones(artifact_dir, lt_endpoint, lt_token, env_key=env_key)

    for app in app_data_list:
        deploy_zone_key = ""
        # Get the target deployment zone based on the name provided in the manifest
        target_deploy_zone = next(filter(lambda x: x["Name"] == app["DeploymentZone"], deploy_zones), None)
        try:
            # Get the status of the app in the target env, to check if they were deployed
            app_status = get_environment_app_version(artifact_dir, lt_endpoint, lt_token, True, env_name=env_name, app_key=app["Key"])
            # Check if the app version is already deployed in the target environment
            for app_in_env in app_status["AppStatusInEnvs"]:
                if app_in_env["EnvironmentKey"] == env_key:
                    # Check if the target environment has the version deployed
                    if app_in_env["BaseApplicationVersionKey"] != app["VersionKey"]:
                        # The version is not the one deployed -> need to compare the version tag
                        app_in_env_data = get_application_version(artifact_dir, lt_endpoint, lt_token, False, app_in_env["BaseApplicationVersionKey"], app_key=app["Key"])
                        # If the version in the target environment has the same version number -> skip deployment
                        if Version(app_in_env_data["Version"]) == Version(app["Version"]):
                            print("Skipping application {} with version {}, since it's already deployed in {} environment.\nReason: VersionTag is equal.".format(app["Name"], app["Version"], env_name), flush=True)
                        else:
                            # Generated app_keys for deployment plan based on the target version
                            if target_deploy_zone:
                                # Check if target deployment zone is different from the current one being used
                                if target_deploy_zone["Key"] != app_in_env["DeploymentZoneKey"]:
                                    deploy_zone_key = target_deploy_zone["Key"]
                            elif include_deployment_zones and app["DeploymentZone"]:
                                print("Deployment zone with name {} not found in {} environment.".format(app["DeploymentZone"], env_name), flush=True)
                            app_keys.append(generate_deploy_app_key(lt_api_version, app["VersionKey"], deploy_zone_key))
                            if deploy_zone_key:
                                print("Adding application {} with version {}, to be deployed in {} environment using {} deployment zone.".format(app["Name"], app["Version"], env_name, target_deploy_zone["Name"]), flush=True)
                            else:
                                print("Adding application {} with version {}, to be deployed in {} environment.".format(app["Name"], app["Version"], env_name), flush=True)
                    else:
                        print("Skipping application {} with version {}, since it's already deployed in {} environment.\nReason: VersionKey is equal.".format(app["Name"], app["Version"], env_name), flush=True)
        except AppDoesNotExistError:
            if target_deploy_zone:
                deploy_zone_key = target_deploy_zone["Key"]
            elif include_deployment_zones and app["DeploymentZone"]:
                print("Deployment zone with name {} not found in {} environment.".format(app["DeploymentZone"], env_name), flush=True)
            app_keys.append(generate_deploy_app_key(lt_api_version, app["VersionKey"], deploy_zone_key))  # type: ignore
            if deploy_zone_key:
                print("App {} with version {} does not exist in {} environment. Ignoring check and deploying it using {} deployment zone.".format(app["Name"], app["Version"], env_name, target_deploy_zone["Name"]), flush=True)
            else:
                print("App {} with version {} does not exist in {} environment. Ignoring check and deploying it.".format(app["Name"], app["Version"], env_name), flush=True)
    return app_keys


def main(artifact_dir: str, lt_http_proto: str, lt_url: str, lt_api_endpoint: str, lt_api_version: int, lt_token: str, source_env_label: str, dest_env_label: str, include_test_apps: bool, trigger_manifest: dict, force_two_step_deployment: bool, include_deployment_zones: bool, allow_parallel_deployments: bool):

    app_data_list = []  # will contain the applications to deploy details from LT
    to_deploy_app_keys = []  # will contain the app keys for the apps tagged

    # Builds the LifeTime endpoint
    lt_endpoint = build_lt_endpoint(lt_http_proto, lt_url, lt_api_endpoint, lt_api_version)

    # Tuple with (EnvName, EnvKey): src_env_tuple[0] = EnvName; src_env_tuple[1] = EnvKey
    src_env_tuple = get_environment_details(trigger_manifest, source_env_label)
    # Tuple with (EnvName, EnvKey): dest_env_tuple[0] = EnvName; dest_env_tuple[1] = EnvKey
    dest_env_tuple = get_environment_details(trigger_manifest, dest_env_label)

    # Retrive the app versions to deploy from the manifest content
    app_data_list = generate_deployment_based_on_manifest(artifact_dir, lt_endpoint, lt_token, src_env_tuple[1], src_env_tuple[0], trigger_manifest, include_test_apps, include_deployment_zones)

    # Check if which application versions have not been deployed to destination environment
    to_deploy_app_keys = check_if_can_deploy(artifact_dir, lt_endpoint, lt_api_version, lt_token, dest_env_tuple[1], dest_env_tuple[0], app_data_list, include_deployment_zones)

    # Check if there are apps to be deployed
    if len(to_deploy_app_keys) == 0:
        print("Deployment skipped because {} environment already has the target application deployed with the same tags.".format(dest_env_tuple[0]), flush=True)
        sys.exit(0)

    # Write the names and keys of the application versions to be deployed
    to_deploy_app_names = []
    to_deploy_app_info = []
    for app in app_data_list:
        for deploying_apps in to_deploy_app_keys:
            if lt_api_version == 1:  # LT for OS version < 11
                if deploying_apps == app["VersionKey"]:
                    to_deploy_app_names.append(app["Name"])
                    to_deploy_app_info.append(app)
            elif lt_api_version == 2:  # LT for OS v11
                if deploying_apps["ApplicationVersionKey"] == app["VersionKey"]:
                    to_deploy_app_names.append(app["Name"])
                    to_deploy_app_info.append(app)
            else:
                raise NotImplementedError("Please make sure the API version is compatible with the module.")
    print("Creating deployment plan from {} (Label: {}) to {} (Label: {}) including applications: {} ({}).".format(src_env_tuple[0], source_env_label, dest_env_tuple[0], dest_env_label, to_deploy_app_names, to_deploy_app_info), flush=True)

    if not allow_parallel_deployments:
        wait_counter = 0
        deployments = get_running_deployment(artifact_dir, lt_endpoint, lt_token, dest_env_tuple[1])
        while len(deployments) > 0:
            if wait_counter >= get_configuration_value("QUEUE_TIMEOUT_IN_SECS", QUEUE_TIMEOUT_IN_SECS):
                print("Timeout occurred while waiting for LifeTime to be free, to create the new deployment plan.", flush=True)
                sys.exit(1)
            sleep_value = get_configuration_value("SLEEP_PERIOD_IN_SECS", SLEEP_PERIOD_IN_SECS)
            sleep(sleep_value)
            wait_counter += sleep_value
            print("Waiting for LifeTime to be free. Elapsed time: {} seconds...".format(wait_counter), flush=True)
            deployments = get_running_deployment(artifact_dir, lt_endpoint, lt_token, dest_env_tuple[1])

    # LT is free to deploy
    # Send the deployment plan and grab the key
    dep_plan_key = send_deployment(artifact_dir, lt_endpoint, lt_token, lt_api_version, to_deploy_app_keys, get_deployment_notes(trigger_manifest), src_env_tuple[0], dest_env_tuple[0])
    print("Deployment plan {} created successfully.".format(dep_plan_key), flush=True)

    # Check if created deployment plan has conflicts
    dep_details = get_deployment_info(artifact_dir, lt_endpoint, lt_token, dep_plan_key)
    has_conflicts = len(dep_details["ApplicationConflicts"]) > 0
    if has_conflicts:
        store_data(artifact_dir, CONFLICTS_FILE, dep_details["ApplicationConflicts"])
        if not get_configuration_value("ALLOW_CONTINUE_WITH_ERRORS", ALLOW_CONTINUE_WITH_ERRORS) or lt_api_version == 1:
            print("Deployment plan {} has conflicts and will be aborted. Check {} artifact for more details.".format(dep_plan_key, CONFLICTS_FILE), flush=True)
            # Abort previously created deployment plan to target environment
            delete_deployment(lt_endpoint, lt_token, dep_plan_key)
            print("Deployment plan {} was deleted successfully.".format(dep_plan_key), flush=True)
            sys.exit(1)
        else:
            print("Deployment plan {} has conflicts but will continue with errors. Check {} artifact for more details.".format(dep_plan_key, CONFLICTS_FILE), flush=True)

    # Check if outdated consumer applications (outside of deployment plan) should be redeployed and start the deployment plan execution
    if lt_api_version == 1:  # LT for OS version < 11
        start_deployment(lt_endpoint, lt_token, dep_plan_key)
    elif lt_api_version == 2:  # LT for OS v11
        if has_conflicts:
            start_deployment(lt_endpoint, lt_token, dep_plan_key, redeploy_outdated=False, continue_with_errors=True)
        else:
            start_deployment(lt_endpoint, lt_token, dep_plan_key, redeploy_outdated=get_configuration_value("REDEPLOY_OUTDATED_APPS", REDEPLOY_OUTDATED_APPS))
    else:
        raise NotImplementedError("Please make sure the API version is compatible with the module.")
    print("Deployment plan {} started being executed.".format(dep_plan_key), flush=True)

    # Flag to only alert the user once
    alert_user = False
    # Sleep thread until deployment has finished
    wait_counter = 0
    while wait_counter < get_configuration_value("DEPLOYMENT_TIMEOUT_IN_SECS", DEPLOYMENT_TIMEOUT_IN_SECS):
        # Check Deployment Plan status.
        dep_status = get_deployment_status(
            artifact_dir, lt_endpoint, lt_token, dep_plan_key)
        if dep_status["DeploymentStatus"] != DEPLOYMENT_RUNNING_STATUS:
            # Check deployment status is pending approval.
            if dep_status["DeploymentStatus"] == DEPLOYMENT_WAITING_STATUS:
                # Check if deployment waiting status is due to 2-Step
                if check_deployment_two_step_deploy_status(dep_status):
                    # Force it to continue in case of force_two_step_deployment parameter
                    if force_two_step_deployment:
                        continue_deployment(lt_endpoint, lt_token, dep_plan_key)
                        print("Deployment plan {} resumed execution.".format(dep_plan_key), flush=True)
                    else:
                        # Exit the script to continue with the pipeline execution
                        print("Deployment plan {} first step finished successfully.".format(dep_plan_key), flush=True)
                        sys.exit(0)
                # Send notification to alert deployment manual intervention.
                elif not alert_user:
                    alert_user = True
                    print("A manual intervention is required to continue the execution of the deployment plan {}.".format(dep_plan_key), flush=True)
            elif dep_status["DeploymentStatus"] in DEPLOYMENT_ERROR_STATUS_LIST:
                print("Deployment plan finished with status {}.".format(dep_status["DeploymentStatus"]), flush=True)
                store_data(artifact_dir, DEPLOY_ERROR_FILE, dep_status)
                sys.exit(1)
            else:
                # If it reaches here, it means the deployment was successful
                print("Deployment plan finished with status {}.".format(dep_status["DeploymentStatus"]), flush=True)
                # Exit the script to continue with the pipeline
                sys.exit(0)
        # Deployment status is still running. Go back to sleep.
        sleep_value = get_configuration_value("SLEEP_PERIOD_IN_SECS", SLEEP_PERIOD_IN_SECS)
        sleep(sleep_value)
        wait_counter += sleep_value
        print("{} secs have passed since the deployment started...".format(wait_counter), flush=True)

    # Deployment timeout reached. Exit script with error
    print("Timeout occurred while deployment plan is still in {} status.".format(DEPLOYMENT_RUNNING_STATUS), flush=True)
    sys.exit(1)


# End of main()


if __name__ == "__main__":
    # Argument menu / parsing
    parser = argparse.ArgumentParser()
    parser.add_argument("-a", "--artifacts", type=str, default=ARTIFACT_FOLDER,
                        help="(Optional) Name of the artifacts folder. Default: \"Artifacts\"")
    parser.add_argument("-u", "--lt_url", type=str, required=True,
                        help="URL for LifeTime environment, without the API endpoint. Example: \"https://<lifetime_host>\"")
    parser.add_argument("-t", "--lt_token", type=str, required=True,
                        help="Token for LifeTime API calls.")
    parser.add_argument("-v", "--lt_api_version", type=int, default=LIFETIME_API_VERSION,
                        help="(Optional) LifeTime API version number. If version <= 10, use 1, if version >= 11, use 2. Default: 2")
    parser.add_argument("-e", "--lt_endpoint", type=str, default=LIFETIME_API_ENDPOINT,
                        help="(Optional) Used to set the API endpoint for LifeTime, without the version. Default: \"lifetimeapi/rest\"")
    parser.add_argument("-s", "--source_env_label", type=str, required=True,
                        help="Label, as configured in the manifest, of the source environment where the apps are.")
    parser.add_argument("-d", "--destination_env_label", type=str, required=True,
                        help="Label, as configured in the manifest, of the destination environment where you want to deploy the apps.")
    parser.add_argument("-i", "--include_test_apps", action='store_true',
                        help="Flag that indicates if applications marked as \"Test Application\" in the manifest are included in the deployment plan.")
    parser.add_argument("-m", "--trigger_manifest", type=str,
                        help="Manifest artifact (in JSON format) received when the pipeline is triggered. Contains required data used throughout the pipeline execution.")
    parser.add_argument("-f", "--manifest_file", type=str,
                        help="Manifest file (with JSON format). Contains required data used throughout the pipeline execution.")
    parser.add_argument("-c", "--force_two_step_deployment", action='store_true',
                        help="Force the execution of the 2-Step deployment.")
    parser.add_argument("-z", "--include_deployment_zones", action='store_true',
                        help="Flag that indicates if deployment zone selection is included in the deployment plan. Applicable to self-managed environments only.")
    parser.add_argument("-cf", "--config_file", type=str,
                        help="Config file path. Contains configuration values to override the default ones.")
    parser.add_argument("-p", "--allow_parallel_deployments", action='store_true',
                        help="Skip LifeTime validation for active deployment plans.")

    args = parser.parse_args()

    # Load config file if exists
    if args.config_file:
        load_configuration_file(args.config_file)
    # Parse the artifact directory
    artifact_dir = args.artifacts
    # Parse the API endpoint
    lt_api_endpoint = args.lt_endpoint
    # Parse the LT Url and split the LT hostname from the HTTP protocol
    # Assumes the default HTTP protocol = https
    lt_http_proto = LIFETIME_HTTP_PROTO
    lt_url = args.lt_url
    if lt_url.startswith("http://"):
        lt_http_proto = "http"
        lt_url = lt_url.replace("http://", "")
    else:
        lt_url = lt_url.replace("https://", "")
    if lt_url.endswith("/"):
        lt_url = lt_url[:-1]
    # Parte LT API Version
    lt_version = args.lt_api_version
    # Parse the LT Token
    lt_token = args.lt_token
    # Parse Source Environment
    source_env_label = args.source_env_label
    # Parse Destination Environment
    dest_env_label = args.destination_env_label
    # Parse Include Test Apps flag
    include_test_apps = args.include_test_apps

    # Validate Manifest is being passed either as JSON or as file
    if not args.trigger_manifest and not args.manifest_file:
        raise ManifestDoesNotExistError("The manifest was not provided as JSON or as a file. Aborting!")

    # Parse Trigger Manifest artifact
    if args.manifest_file:
        trigger_manifest_path = os.path.split(args.manifest_file)
        trigger_manifest = load_data(trigger_manifest_path[0], trigger_manifest_path[1])
    else:
        trigger_manifest = json.loads(args.trigger_manifest)

    # Parse Force Two-step Deployment flag
    force_two_step_deployment = args.force_two_step_deployment
    # Parse Include Deployment Zones flag
    include_deployment_zones = args.include_deployment_zones

    # Parse Allow Parallel Deployments
    allow_parallel_deployments = args.allow_parallel_deployments

    # Calls the main script
    main(artifact_dir, lt_http_proto, lt_url, lt_api_endpoint, lt_version, lt_token, source_env_label, dest_env_label, include_test_apps, trigger_manifest, force_two_step_deployment, include_deployment_zones, allow_parallel_deployments)



# Python Modules
import os

# Custom Modules
# Functions
from outsystems.cicd_probe.cicd_base import send_probe_get_request
from outsystems.file_helpers.file import store_data
# Variables
from outsystems.vars.cicd_vars import SCAN_BDD_TESTS_ENDPOINT, PROBE_SCAN_SUCCESS_CODE, PROBE_API_VERSION
from outsystems.vars.file_vars import PROBE_APPLICATION_SCAN_FILE, PROBE_FOLDER


# Scan existing BDD test endpoints (i.e. Screens) in the target environment.
def scan_bdd_test_endpoint(artifact_dir: str, endpoint: str, application_name: str, api_key: str = None,
                           exclude_pattern: str = None, probe_version: int = PROBE_API_VERSION):
    # Builds the API params
    params = {"ApplicationName": application_name}
    if exclude_pattern:
        if probe_version == 1:
            params["WebFlowExcludePattern"] = exclude_pattern
        elif probe_version == 2:
            params["TestFlowExcludePattern"] = exclude_pattern
    # Sends the request
    response = send_probe_get_request(
        endpoint, SCAN_BDD_TESTS_ENDPOINT, api_key, params)
    status_code = response["http_status"]
    if status_code == PROBE_SCAN_SUCCESS_CODE:
        # Stores the result
        filename = "{}{}".format(application_name, PROBE_APPLICATION_SCAN_FILE)
        filename = os.path.join(PROBE_FOLDER, filename)
        store_data(artifact_dir, filename, response["response"])
        return response["response"]
    else:
        raise NotImplementedError(
            "There was an error. Response from server: {}".format(response))



# Python Modules
import sys
import os
import argparse

# Workaround for Jenkins:
# Set the path to include the outsystems module
# Jenkins exposes the workspace directory through env.
if "WORKSPACE" in os.environ:
    sys.path.append(os.environ['WORKSPACE'])
else:  # Else just add the project dir
    sys.path.append(os.getcwd())

# Custom Modules
# Variables
from outsystems.vars.file_vars import ARTIFACT_FOLDER
from outsystems.vars.lifetime_vars import LIFETIME_HTTP_PROTO, LIFETIME_API_ENDPOINT, LIFETIME_API_VERSION
from outsystems.vars.pipeline_vars import MAX_VERSIONS_TO_RETURN, TAG_APP_MAX_RETRIES
from outsystems.vars.manifest_vars import MANIFEST_APPLICATION_VERSIONS

# Functions
from outsystems.lifetime.lifetime_applications import get_applications, get_running_app_version, get_application_versions
from outsystems.lifetime.lifetime_environments import get_environment_key
from outsystems.lifetime.lifetime_base import build_lt_endpoint
from outsystems.lifetime.lifetime_applications import set_application_version
from outsystems.file_helpers.file import load_data
from outsystems.vars.vars_base import load_configuration_file, get_configuration_value

# Exceptions
from outsystems.exceptions.invalid_parameters import InvalidParametersError


# ############################################################# SCRIPT ##############################################################
def generate_new_version_number(base_version: str):
    # Split tag version digits
    base_version = base_version.split('.')  # type: ignore

    # Default values for Major.Minor.Revison
    maj = base_version[0]
    min = "0"
    rev = "1"

    if not len(base_version) < 2:
        min = base_version[1]

    if not len(base_version) < 3:
        # Increment revision number
        rev = str(int(base_version[2]) + 1)

    return "{}.{}.{}".format(maj, min, rev)


def main(artifact_dir: str, lt_http_proto: str, lt_url: str, lt_api_endpoint: str, lt_api_version: int, lt_token: str, dest_env: str, apps: list, trigger_manifest: dict, log_msg: str):
    # Builds the LifeTime endpoint
    lt_endpoint = build_lt_endpoint(lt_http_proto, lt_url, lt_api_endpoint, lt_api_version)

    # Get the environment key
    env_key = get_environment_key(artifact_dir, lt_endpoint, lt_token, dest_env)

    # Get all applications info
    all_apps = get_applications(artifact_dir, lt_endpoint, lt_token, True)

    # Use trigger_manifest or apps list
    app_list = trigger_manifest[MANIFEST_APPLICATION_VERSIONS] if trigger_manifest else apps
    trigger_in_use = bool(trigger_manifest)

    for app in app_list:

        app_name = app["ApplicationName"] if trigger_in_use else app
        # Gets application specific details
        app_detail = list(filter(lambda x: x["Name"] == app_name, all_apps))

        if len(app_detail):
            # Checks if application is modified in target env
            app_env_detail = list(filter(lambda x: x["EnvironmentKey"] == env_key and x["IsModified"], app_detail[0]["AppStatusInEnvs"]))

            if len(app_env_detail):
                current_tag = get_running_app_version(artifact_dir, lt_endpoint, lt_token, env_key, app_name=app_name)
                generated_tag = generate_new_version_number(current_tag["Version"])

                # List of the last application tags
                tag_history_list = [d["Version"] for d in get_application_versions(artifact_dir, lt_endpoint, lt_token, get_configuration_value("MAX_VERSIONS_TO_RETURN", MAX_VERSIONS_TO_RETURN), app_name=app_name)]

                # Finds next available tag number
                retries = 0
                while retries < get_configuration_value("TAG_APP_MAX_RETRIES", TAG_APP_MAX_RETRIES):
                    if generated_tag in tag_history_list:
                        generated_tag = generate_new_version_number(generated_tag)
                    else:
                        # Checks if app is mobile and gets mobile info
                        app_mobile_detail = list(filter(lambda x: x["IsModified"], app_env_detail[0]["MobileAppsStatus"]))

                        # Will contain the List of mobile versions to tag
                        native_shell_versions = []

                        # Generate new version number for each native shell
                        if app_mobile_detail:
                            for native_shell in app_mobile_detail:
                                native_shell_versions.append({"NativePlatform": native_shell["NativePlatform"], "VersionNumber": generate_new_version_number(native_shell["VersionNumber"]), "VersionDescription": log_msg})

                        set_application_version(lt_endpoint, lt_token, env_key, current_tag["ApplicationKey"], log_msg, generated_tag, native_shell_versions)
                        print("Application '{}' successfully tagged to version {} on environment '{}'".format(current_tag["ApplicationName"], generated_tag, dest_env), flush=True)
                        break

                    retries += 1
                    if retries == get_configuration_value("TAG_APP_MAX_RETRIES", TAG_APP_MAX_RETRIES):
                        print("Could not find available tag for Application '{}' ".format(current_tag["ApplicationName"]), flush=True)

# End of main()


if __name__ == "__main__":
    # Argument menu / parsing
    parser = argparse.ArgumentParser()
    parser.add_argument("-a", "--artifacts", type=str, default=ARTIFACT_FOLDER,
                        help="Name of the artifacts folder. Default: \"Artifacts\"")
    parser.add_argument("-u", "--lt_url", type=str, required=True,
                        help="URL for LifeTime environment, without the API endpoint. Example: \"https://<lifetime_host>\"")
    parser.add_argument("-t", "--lt_token", type=str, required=True,
                        help="Token for LifeTime API calls.")
    parser.add_argument("-v", "--lt_api_version", type=int, default=LIFETIME_API_VERSION,
                        help="LifeTime API version number. If version <= 10, use 1, if version >= 11, use 2. Default: 2")
    parser.add_argument("-e", "--lt_endpoint", type=str, default=LIFETIME_API_ENDPOINT,
                        help="(optional) Used to set the API endpoint for LifeTime, without the version. Default: \"lifetimeapi/rest\"")
    parser.add_argument("-d", "--destination_env", type=str, required=True,
                        help="Name, as displayed in LifeTime, of the environment where you want to tag the apps.")
    parser.add_argument("-l", "--app_list", type=str,
                        help="Comma separated list of apps you want to tag. Example: \"App1,App2 With Spaces,App3_With_Underscores\"")
    parser.add_argument("-f", "--manifest_file", type=str,
                        help="Manifest file (with JSON format). Contains required data used throughout the pipeline execution.")
    parser.add_argument("-lm", "--log_msg", type=str, default="Version created automatically using outsystems-pipeline package.",
                        help="(optional) log message to be added to the new tags")
    parser.add_argument("-cf", "--config_file", type=str,
                        help="(optional) Config file path. Contains configuration values to override the default ones.")

    args = parser.parse_args()

    # Load config file if exists
    if args.config_file:
        load_configuration_file(args.config_file)
    # Parse the artifact directory
    artifact_dir = args.artifacts
    # Parse the API endpoint
    lt_api_endpoint = args.lt_endpoint
    # Parse the LT Url and split the LT hostname from the HTTP protocol
    # Assumes the default HTTP protocol = https
    lt_http_proto = LIFETIME_HTTP_PROTO
    lt_url = args.lt_url
    if lt_url.startswith("http://"):
        lt_http_proto = "http"
        lt_url = lt_url.replace("http://", "")
    else:
        lt_url = lt_url.replace("https://", "")
    if lt_url.endswith("/"):
        lt_url = lt_url[:-1]
    # Parte LT API Version
    lt_version = args.lt_api_version
    # Parse the LT Token
    lt_token = args.lt_token
    # Parse Destination Environment
    dest_env = args.destination_env
    # Parse App list (if it exists)
    apps = None
    if args.app_list:
        apps = args.app_list.split(',')
    # Parse Manifest file (if it exists)
    trigger_manifest = None
    if args.manifest_file:
        manifest_path = os.path.split(args.manifest_file)
        trigger_manifest = load_data(manifest_path[0], manifest_path[1])
    # Check if either an app list or a manifest file is being provided
    if not args.app_list and not args.manifest_file:
        raise InvalidParametersError("either --app_list or --manifest_file must be provided as arguments")

    # Parse Log Message
    log_msg = args.log_msg
    # Calls the main script
    main(artifact_dir, lt_http_proto, lt_url, lt_api_endpoint, lt_version, lt_token, dest_env, apps, trigger_manifest, log_msg)



# Directory Vars
ARTIFACT_FOLDER = "Artifacts"

# Applications vars
APPLICATIONS_FILE = "applications.cache"
APPLICATION_FILE = ".cache"
APPLICATION_VERSIONS_FILE = ".versions.cache"
APPLICATION_VERSION_FILE = ".version.cache"
APPLICATION_FOLDER = "application_data"
APPLICATION_OAP_FOLDER = "application_oap"
APPLICATION_OAP_FILE = ".oap"

# Modules vars
MODULES_FILE = "modules.cache"
MODULES_FOLDER = "module_data"
MODULE_FILE = ".cache"
MODULE_VERSIONS_FILE = ".versions.cache"
MODULE_VERSION_FILE = ".version.cache"

# Environments vars
ENVIRONMENTS_FILE = "environments.cache"
ENVIRONMENT_FILE = ".cache"
ENVIRONMENT_APPS_FILE = ".applications.cache"
ENVIRONMENT_APPLICATION_FILE = ".applications.cache"
ENVIRONMENT_VERSIONS_FILE = ".versions.cache"
ENVIRONMENT_DEPLOYMENT_ZONES_FILE = ".deploymentzones.cache"
ENVIRONMENT_FOLDER = "environment_data"
ENVIRONMENT_SOURCECODE_FOLDER = "sourcecode_data"
ENVIRONMENT_SOURCECODE_STATUS_FILE = ".status.cache"
ENVIRONMENT_SOURCECODE_LINK_FILE = ".link.cache"
ENVIRONMENT_SOURCECODE_DOWNLOAD_FILE = ".source.zip"

# Deployments vars
DEPLOYMENTS_FILE = "deployments.cache"
DEPLOYMENT_FILE = ".cache"
DEPLOYMENT_PLAN_FILE = ".plan.cache"
DEPLOYMENT_MANIFEST_FILE = "deployment_manifest.cache"
DEPLOYMENT_STATUS_FILE = ".status.cache"
DEPLOYMENT_FOLDER = "deployment_data"

# CICD Probe vars
PROBE_APPLICATION_SCAN_FILE = ".probe.cache"
PROBE_APPLICATION_DEPENDENCIES_FILE = ".dependencies.cache"
PROBE_FOLDER = "cicd_probe_data"

# BDD Framework vars
BDD_FRAMEWORK_TEST_RUN_FILE = ".testrun.cache"
BDD_FRAMEWORK_TEST_ENDPOINTS_FILE = "test.endpoints.cache"
BDD_FRAMEWORK_FOLDER = "bdd_data"

# JUNIT vars
JUNIT_TEST_RESULTS_FILE = "junit-result.xml"

# Architecture Dashboard vars
AD_FILE_PREFIX = "TechDebt"
AD_LEVELS_FILE = ".levels.cache"
AD_CATEGORIES_FILE = ".categories.cache"
AD_INFRA_FILE = ".infrastructure.cache"
AD_APP_FILE = ".application.cache"
AD_FOLDER = "techdebt_data"

# AirGap vars
DEPLOYMENT_ORDER_FILE = "sorted_oap.list"

# Solutions vars
SOLUTIONS_OSP_FILE = ".osp"
SOLUTIONS_LINK_FILE = ".link.cache"
SOLUTIONS_STATUS_FILE = ".status.cache"
SOLUTIONS_DEPLOY_FILE = ".deploy.cache"
SOLUTIONS_FOLDER = "solution_data"



# Python Modules
import json
import os


def store_data(artifact_dir: str, filename: str, data: str):
    filename = os.path.join(artifact_dir, filename)
    # Remove the spaces in the filename
    filename = filename.replace(" ", "_")
    # Makes sure that, if a directory is in the filename, that directory exists
    os.makedirs(os.path.dirname(filename), exist_ok=True)
    with open(filename, "w") as outfile:
        json.dump(data, outfile, indent=4)


def load_data(artifact_dir: str, filename: str):
    # Remove the spaces in the filename
    filename = filename.replace(" ", "_")
    if check_file(artifact_dir, filename):
        filename = os.path.join(artifact_dir, filename)
        with open(filename, "r") as infile:
            return json.load(infile)
    raise FileNotFoundError(
        "The file with filename {} does not exist.".format(filename))


def check_file(artifact_dir: str, filename: str):
    filename = os.path.join(artifact_dir, filename)
    return os.path.isfile(filename)


def clear_cache(artifact_dir: str, filename: str):
    if not check_file(artifact_dir, filename):
        return
    filename = os.path.join(artifact_dir, filename)
    os.remove(filename)


# Returns a human readable string representation of bytes
def bytes_human_readable_size(bytes, units=[' bytes', 'KB', 'MB', 'GB', 'TB', 'PB', 'EB']):
    return str(bytes) + units[0] if bytes < 1024 else bytes_human_readable_size(bytes >> 10, units[1:])


def is_valid_os_package(filename: str):
    return filename.lower().split('.')[-1] in ("osp", "oap")



# Python Modules
import sys
import os
import argparse
from time import sleep

# Workaround for Jenkins:
# Set the path to include the outsystems module
# Jenkins exposes the workspace directory through env.
if "WORKSPACE" in os.environ:
    sys.path.append(os.environ['WORKSPACE'])
else:  # Else just add the project dir
    sys.path.append(os.getcwd())

# Custom Modules
# Variables
from outsystems.vars.file_vars import ARTIFACT_FOLDER, DEPLOYMENT_FOLDER, DEPLOYMENT_MANIFEST_FILE
from outsystems.vars.lifetime_vars import LIFETIME_HTTP_PROTO, LIFETIME_API_ENDPOINT, LIFETIME_API_VERSION
from outsystems.vars.pipeline_vars import SLEEP_PERIOD_IN_SECS, CONFLICTS_FILE, REDEPLOY_OUTDATED_APPS, \
    DEPLOYMENT_TIMEOUT_IN_SECS, DEPLOYMENT_RUNNING_STATUS, DEPLOYMENT_WAITING_STATUS, \
    DEPLOYMENT_ERROR_STATUS_LIST, DEPLOY_ERROR_FILE
# Functions
from outsystems.lifetime.lifetime_environments import get_environment_key
from outsystems.lifetime.lifetime_applications import get_application_version, get_application_data
from outsystems.lifetime.lifetime_deployments import get_deployment_status, get_deployment_info, \
    start_deployment, continue_deployment, get_saved_deployment
from outsystems.file_helpers.file import store_data
from outsystems.lifetime.lifetime_base import build_lt_endpoint
from outsystems.vars.vars_base import get_configuration_value, load_configuration_file
# Exceptions
from outsystems.exceptions.deployment_not_found import DeploymentNotFoundError

# ############################################################# SCRIPT ##############################################################


# Function that will create a manifest file based on the deployment provided as input
def generate_deployment_manifest(artifact_dir: str, lt_endpoint: str, lt_token: str, deployment: dict):
    app_operations_list = []  # will contain the applications to deploy details from LT plan
    deployment_manifest = []  # will store the deployment manifest, that may be used in later stages of the pipeline

    # Get list of applications to deploy from deployment plan
    # dep_plan = deployment["Deployment"]
    if "ApplicationOperations" in deployment:
        app_operations_list = deployment["ApplicationOperations"]

    # Creates a list with the details for the apps to deploy
    for app_operation in app_operations_list:
        # Get details from the application to deploy
        app_to_deploy = get_application_data(artifact_dir, lt_endpoint, lt_token, False, app_key=app_operation["ApplicationKey"])

        # Get details from the base version to deploy
        base_version_to_deploy = get_application_version(artifact_dir, lt_endpoint, lt_token, False, app_operation["ApplicationVersionKey"], app_key=app_operation["ApplicationKey"])

        # Create manifest file entry
        entry = {'Name': app_to_deploy["Name"], 'Key': app_to_deploy["Key"], 'Version': base_version_to_deploy["Version"], 'VersionKey': base_version_to_deploy["Key"]}

        # Since these 2 fields were only introduced in a minor of OS11, we check here if they exist
        if "CreatedOn" in base_version_to_deploy:
            entry.update({"CreatedOn": base_version_to_deploy["CreatedOn"]})
        if "ChangeLog" in base_version_to_deploy:
            entry.update({"ChangeLog": base_version_to_deploy["ChangeLog"]})

        # Add deployment operation to the manifest file (in case of Tag & Deploy actions)
        entry.update({"DeploymentOperation": app_operation["DeploymentOperation"]})

        # Add entry to manifest file
        deployment_manifest.append(entry)

    # Store the manifest to be used in other stages of the pipeline
    filename = "{}/{}".format(DEPLOYMENT_FOLDER, DEPLOYMENT_MANIFEST_FILE)
    store_data(ARTIFACT_FOLDER, filename, deployment_manifest)


def main(artifact_dir: str, lt_http_proto: str, lt_url: str, lt_api_endpoint: str, lt_api_version: int, lt_token: str, dest_env: str):

    # Builds the LifeTime endpoint
    lt_endpoint = build_lt_endpoint(lt_http_proto, lt_url, lt_api_endpoint, lt_api_version)

    # Gets the environment key for the destination environment
    dest_env_key = get_environment_key(artifact_dir, lt_endpoint, lt_token, dest_env)

    # Find deployment plan with 'saved' status in destination environment
    deployment = get_saved_deployment(artifact_dir, lt_endpoint, lt_token, dest_env_key)
    if deployment is None:
        raise DeploymentNotFoundError("Unable to find a created deployment plan for {} environment.".format(dest_env))

    # Grab the key from the deployment plan found
    dep_plan_key = deployment["Key"]
    print("Deployment plan {} was found.".format(dep_plan_key), flush=True)

    # Check if created deployment plan has conflicts
    dep_details = get_deployment_info(artifact_dir, lt_endpoint, lt_token, dep_plan_key)
    if len(dep_details["ApplicationConflicts"]) > 0:
        store_data(artifact_dir, CONFLICTS_FILE, dep_details["ApplicationConflicts"])
        print("Deployment plan {} has conflicts and will be aborted. Check {} artifact for more details.".format(dep_plan_key, CONFLICTS_FILE), flush=True)
        # Previously created deployment plan to target environment will NOT be deleted
        sys.exit(1)

    # Generate deployment manifest based on content of custom deployment plan
    generate_deployment_manifest(artifact_dir, lt_endpoint, lt_token, deployment)

    # Check if outdated consumer applications (outside of deployment plan) should be redeployed and start the deployment plan execution
    if lt_api_version == 1:  # LT for OS version < 11
        start_deployment(lt_endpoint, lt_token, dep_plan_key)
    elif lt_api_version == 2:  # LT for OS v11
        start_deployment(lt_endpoint, lt_token, dep_plan_key, redeploy_outdated=REDEPLOY_OUTDATED_APPS)
    else:
        raise NotImplementedError("Please make sure the API version is compatible with the module.")
    print("Deployment plan {} started being executed.".format(dep_plan_key), flush=True)

    # Sleep thread until deployment has finished
    wait_counter = 0
    while wait_counter < get_configuration_value("DEPLOYMENT_TIMEOUT_IN_SECS", DEPLOYMENT_TIMEOUT_IN_SECS):
        # Check Deployment Plan status.
        dep_status = get_deployment_status(
            artifact_dir, lt_endpoint, lt_token, dep_plan_key)
        if dep_status["DeploymentStatus"] != DEPLOYMENT_RUNNING_STATUS:
            # Check deployment status is pending approval. Force it to continue (if 2-Step deployment is enabled)
            if dep_status["DeploymentStatus"] == DEPLOYMENT_WAITING_STATUS:
                continue_deployment(lt_endpoint, lt_token, dep_plan_key)
                print("Deployment plan {} resumed execution.".format(dep_plan_key), flush=True)
            elif dep_status["DeploymentStatus"] in DEPLOYMENT_ERROR_STATUS_LIST:
                print("Deployment plan finished with status {}.".format(dep_status["DeploymentStatus"]), flush=True)
                store_data(artifact_dir, DEPLOY_ERROR_FILE, dep_status)
                sys.exit(1)
            else:
                # If it reaches here, it means the deployment was successful
                print("Deployment plan finished with status {}.".format(dep_status["DeploymentStatus"]), flush=True)
                # Exit the script to continue with the pipeline
                sys.exit(0)
        # Deployment status is still running. Go back to sleep.
        sleep_value = get_configuration_value("SLEEP_PERIOD_IN_SECS", SLEEP_PERIOD_IN_SECS)
        sleep(sleep_value)
        wait_counter += sleep_value
        print("{} secs have passed since the deployment started...".format(wait_counter), flush=True)

    # Deployment timeout reached. Exit script with error
    print("Timeout occurred while deployment plan is still in {} status.".format(DEPLOYMENT_RUNNING_STATUS), flush=True)
    sys.exit(1)


# End of main()


if __name__ == "__main__":
    # Argument menu / parsing
    parser = argparse.ArgumentParser()
    parser.add_argument("-a", "--artifacts", type=str, default=ARTIFACT_FOLDER,
                        help="Name of the artifacts folder. Default: \"Artifacts\"")
    parser.add_argument("-u", "--lt_url", type=str, required=True,
                        help="URL for LifeTime environment, without the API endpoint. Example: \"https://<lifetime_host>\"")
    parser.add_argument("-t", "--lt_token", type=str, required=True,
                        help="Token for LifeTime API calls.")
    parser.add_argument("-v", "--lt_api_version", type=int, default=LIFETIME_API_VERSION,
                        help="LifeTime API version number. If version <= 10, use 1, if version >= 11, use 2. Default: 2")
    parser.add_argument("-e", "--lt_endpoint", type=str, default=LIFETIME_API_ENDPOINT,
                        help="(optional) Used to set the API endpoint for LifeTime, without the version. Default: \"lifetimeapi/rest\"")
    parser.add_argument("-d", "--destination_env", type=str, required=True,
                        help="Name, as displayed in LifeTime, of the destination environment where you want to start the deployment plan.")
    parser.add_argument("-cf", "--config_file", type=str,
                        help="Config file path. Contains configuration values to override the default ones.")

    args = parser.parse_args()

    # Load config file if exists
    if args.config_file:
        load_configuration_file(args.config_file)
    # Parse the artifact directory
    artifact_dir = args.artifacts
    # Parse the API endpoint
    lt_api_endpoint = args.lt_endpoint
    # Parse the LT Url and split the LT hostname from the HTTP protocol
    # Assumes the default HTTP protocol = https
    lt_http_proto = LIFETIME_HTTP_PROTO
    lt_url = args.lt_url
    if lt_url.startswith("http://"):
        lt_http_proto = "http"
        lt_url = lt_url.replace("http://", "")
    else:
        lt_url = lt_url.replace("https://", "")
    if lt_url.endswith("/"):
        lt_url = lt_url[:-1]
    # Parte LT API Version
    lt_version = args.lt_api_version
    # Parse the LT Token
    lt_token = args.lt_token
    # Parse Destination Environment
    dest_env = args.destination_env

    # Calls the main script
    main(artifact_dir, lt_http_proto, lt_url, lt_api_endpoint, lt_version, lt_token, dest_env)



# Python Modules
import sys
import os
import argparse
import json

# Workaround for Jenkins:
# Set the path to include the outsystems module
# Jenkins exposes the workspace directory through env.
if "WORKSPACE" in os.environ:
    sys.path.append(os.environ['WORKSPACE'])
else:  # Else just add the project dir
    sys.path.append(os.getcwd())

# Custom Modules
# Variables
from outsystems.vars.manifest_vars import MANIFEST_CONFIG_ITEM_TYPE, MANIFEST_MODULE_KEY, MANIFEST_CONFIG_ITEM_KEY, \
    MANIFEST_CONFIG_ITEM_TARGET_VALUE, MANIFEST_CONFIG_ITEM_NAME
from outsystems.vars.properties_vars import PROPERTY_TYPE_SITE_PROPERTY, PROPERTY_TYPE_REST_ENDPOINT, PROPERTY_TYPE_SOAP_ENDPOINT, \
    PROPERTY_TYPE_TIMER_SCHEDULE
from outsystems.vars.lifetime_vars import LIFETIME_HTTP_PROTO
from outsystems.vars.file_vars import ARTIFACT_FOLDER
# Functions
from outsystems.file_helpers.file import load_data
from outsystems.manifest.manifest_base import get_configuration_items_for_environment
from outsystems.manifest.manifest_base import get_environment_details
from outsystems.properties.properties_set_value import set_site_property_value, set_rest_endpoint_url, set_soap_endpoint_url, \
    set_timer_schedule
from outsystems.vars.vars_base import load_configuration_file
# Exceptions
from outsystems.exceptions.manifest_does_not_exist import ManifestDoesNotExistError


# Function to apply configuration values to a target environment
def main(artifact_dir: str, lt_http_proto: str, lt_url: str, lt_token: str, target_env_label: str, trigger_manifest: dict):

    # Tuple with (EnvName, EnvKey): target_env_tuple[0] = EnvName; target_env_tuple[1] = EnvKey
    target_env_tuple = get_environment_details(trigger_manifest, target_env_label)

    # Get configuration items defined in the manifest for target environment
    config_items = get_configuration_items_for_environment(trigger_manifest, target_env_tuple[1])

    # Check if there are any configuration item values to apply for target environment
    if len(config_items) == 0:
        print("No configuration item values were found in the manifest for {} (Label: {}).".format(target_env_tuple[0], target_env_label), flush=True)
    else:
        print("Applying new values to configuration items in {} (Label: {})...".format(target_env_tuple[0], target_env_label), flush=True)

    # Apply target value for each configuration item according to its type
    for cfg_item in config_items:
        result = {}
        if cfg_item[MANIFEST_CONFIG_ITEM_TYPE] == PROPERTY_TYPE_SITE_PROPERTY:
            result = set_site_property_value(
                lt_url, lt_token, cfg_item[MANIFEST_MODULE_KEY], target_env_tuple[1], cfg_item[MANIFEST_CONFIG_ITEM_KEY], cfg_item[MANIFEST_CONFIG_ITEM_TARGET_VALUE])
        elif cfg_item[MANIFEST_CONFIG_ITEM_TYPE] == PROPERTY_TYPE_REST_ENDPOINT:
            result = set_rest_endpoint_url(
                lt_url, lt_token, cfg_item[MANIFEST_MODULE_KEY], target_env_tuple[1], cfg_item[MANIFEST_CONFIG_ITEM_KEY], cfg_item[MANIFEST_CONFIG_ITEM_TARGET_VALUE])
        elif cfg_item[MANIFEST_CONFIG_ITEM_TYPE] == PROPERTY_TYPE_SOAP_ENDPOINT:
            result = set_soap_endpoint_url(
                lt_url, lt_token, cfg_item[MANIFEST_MODULE_KEY], target_env_tuple[1], cfg_item[MANIFEST_CONFIG_ITEM_KEY], cfg_item[MANIFEST_CONFIG_ITEM_TARGET_VALUE])
        elif cfg_item[MANIFEST_CONFIG_ITEM_TYPE] == PROPERTY_TYPE_TIMER_SCHEDULE:
            result = set_timer_schedule(
                lt_url, lt_token, cfg_item[MANIFEST_MODULE_KEY], target_env_tuple[1], cfg_item[MANIFEST_CONFIG_ITEM_KEY], cfg_item[MANIFEST_CONFIG_ITEM_TARGET_VALUE])
        else:
            raise NotImplementedError("Configuration item type '{}' not supported.".format(cfg_item[MANIFEST_CONFIG_ITEM_TYPE]))

        # Check returned result after setting configuration item value
        if "Success" in result and result["Success"]:
            print("New value successfully applied to configuration item '{}' ({}).".format(cfg_item[MANIFEST_CONFIG_ITEM_NAME], cfg_item[MANIFEST_CONFIG_ITEM_TYPE]), flush=True)
        else:
            print("Unable to apply new value to configuration item '{}' ({}).\nReason: {}".format(cfg_item[MANIFEST_CONFIG_ITEM_NAME], cfg_item[MANIFEST_CONFIG_ITEM_TYPE], result["Message"]), flush=True)

    # Exit the script to continue with the pipeline
    sys.exit(0)


# End of main()


if __name__ == "__main__":
    # Argument menu / parsing
    parser = argparse.ArgumentParser()
    parser.add_argument("-a", "--artifacts", type=str, default=ARTIFACT_FOLDER,
                        help="(Optional) Name of the artifacts folder. Default: \"Artifacts\"")
    parser.add_argument("-u", "--lt_url", type=str, required=True,
                        help="URL for LifeTime environment, without the API endpoint. Example: \"https://<lifetime_host>\"")
    parser.add_argument("-t", "--lt_token", type=str, required=True,
                        help="Service account token for Properties API calls.")
    parser.add_argument("-e", "--target_env_label", type=str, required=True,
                        help="Label, as configured in the manifest, of the target environment where the configuration values will be applied.")
    parser.add_argument("-m", "--trigger_manifest", type=str,
                        help="Manifest artifact (in JSON format) received when the pipeline is triggered. Contains required data used throughout the pipeline execution.")
    parser.add_argument("-f", "--manifest_file", type=str,
                        help="Manifest file (with JSON format). Contains required data used throughout the pipeline execution.")
    parser.add_argument("-cf", "--config_file", type=str,
                        help="Config file path. Contains configuration values to override the default ones.")

    args = parser.parse_args()

    # Load config file if exists
    if args.config_file:
        load_configuration_file(args.config_file)
    # Parse the artifact directory
    artifact_dir = args.artifacts
    # Parse the LT Url and split the LT hostname from the HTTP protocol
    # Assumes the default HTTP protocol = https
    lt_http_proto = LIFETIME_HTTP_PROTO
    lt_url = args.lt_url
    if lt_url.startswith("http://"):
        lt_http_proto = "http"
        lt_url = lt_url.replace("http://", "")
    else:
        lt_url = lt_url.replace("https://", "")
    if lt_url.endswith("/"):
        lt_url = lt_url[:-1]
    # Parse the LT Token
    lt_token = args.lt_token
    # Parse Destination Environment
    target_env_label = args.target_env_label

    # Validate Manifest is being passed either as JSON or as file
    if not args.trigger_manifest and not args.manifest_file:
        raise ManifestDoesNotExistError("The manifest was not provided as JSON or as a file. Aborting!")

    # Parse Trigger Manifest artifact
    if args.manifest_file:
        trigger_manifest_path = os.path.split(args.manifest_file)
        trigger_manifest = load_data(trigger_manifest_path[0], trigger_manifest_path[1])
    else:
        trigger_manifest = json.loads(args.trigger_manifest)

    # Calls the main script
    main(artifact_dir, lt_http_proto, lt_url, lt_token, target_env_label, trigger_manifest)



# Python Modules
import os
import sys
import argparse

# Workaround for Jenkins:
# Set the path to include the outsystems module
# Jenkins exposes the workspace directory through env.
if "WORKSPACE" in os.environ:
    sys.path.append(os.environ['WORKSPACE'])
else:  # Else just add the project dir
    sys.path.append(os.getcwd())

# Custom Modules
from outsystems.lifetime.lifetime_applications import get_applications
from outsystems.lifetime.lifetime_environments import get_environments
from outsystems.lifetime.lifetime_base import build_lt_endpoint
from outsystems.vars.lifetime_vars import LIFETIME_HTTP_PROTO, LIFETIME_API_ENDPOINT, LIFETIME_API_VERSION
from outsystems.vars.file_vars import ARTIFACT_FOLDER
from outsystems.vars.vars_base import load_configuration_file


# ---------------------- SCRIPT ----------------------
def main(artifact_dir: str, lt_http_proto: str, lt_url: str, lt_api_endpoint: str, lt_api_version: int, lt_token: str):
    # Builds the LifeTime endpoint
    lt_endpoint = build_lt_endpoint(
        lt_http_proto, lt_url, lt_api_endpoint, lt_api_version)

    # Get Environments
    get_environments(artifact_dir, lt_endpoint, lt_token)
    print("OS Environments data retrieved successfully.", flush=True)
    # Get Applications without extra data
    get_applications(artifact_dir, lt_endpoint, lt_token, False)
    print("OS Applications data retrieved successfully.", flush=True)


if __name__ == "__main__":
    # Argument menu / parsing
    parser = argparse.ArgumentParser()
    parser.add_argument("-a", "--artifacts", type=str,
                        help="Name of the artifacts folder. Default: \"Artifacts\"", default=ARTIFACT_FOLDER)
    parser.add_argument("-u", "--lt_url", type=str,
                        help="URL for LifeTime environment, without the API endpoint. Example: \"https://<lifetime_host>\"", required=True)
    parser.add_argument("-t", "--lt_token", type=str,
                        help="Token for LifeTime API calls.", required=True)
    parser.add_argument("-v", "--lt_api_version", type=int,
                        help="LifeTime API version number. If version <= 10, use 1, if version >= 11, use 2. Default: 2", default=LIFETIME_API_VERSION)
    parser.add_argument("-e", "--lt_endpoint", type=str,
                        help="(optional) Used to set the API endpoint for LifeTime, without the version. Default: \"lifetimeapi/rest\"", default=LIFETIME_API_ENDPOINT)
    parser.add_argument("-cf", "--config_file", type=str,
                        help="Config file path. Contains configuration values to override the default ones.")

    args = parser.parse_args()

    # Load config file if exists
    if args.config_file:
        load_configuration_file(args.config_file)
    # Parse the artifact directory
    artifact_dir = args.artifacts
    # Parse the API endpoint
    lt_api_endpoint = args.lt_endpoint
    # Parse the LT Url and split the LT hostname from the HTTP protocol
    # Assumes the default HTTP protocol = https
    lt_http_proto = LIFETIME_HTTP_PROTO
    lt_url = args.lt_url
    if lt_url.startswith("http://"):
        lt_http_proto = "http"
        lt_url = lt_url.replace("http://", "")
    else:
        lt_url = lt_url.replace("https://", "")
    if lt_url.endswith("/"):
        lt_url = lt_url[:-1]
    # Parte LT API Version
    lt_version = args.lt_api_version
    # Parse the LT Token
    lt_token = args.lt_token

    # Calls the main script
    main(artifact_dir, lt_http_proto, lt_url,
         lt_api_endpoint, lt_version, lt_token)



# Base CICD Probe Variables
BDD_HTTP_PROTO = "https"
BDD_API_ENDPOINT = "BDDFramework/rest"
BDD_CLIENT_API_ENDPOINT = "TestRunner_API/rest"
BDD_API_VERSION = 1
BDD_API_SSL_CERT_VERIFY = True
BDD_FRAMEWORK_TYPE_SERVER = "server"
BDD_FRAMEWORK_TYPE_CLIENT = "client"

# Test Runner Endpoint Variables
BDD_TEST_RUNNER_ENDPOINT = "BDDTestRunner"
BDD_RUNNER_SUCCESS_CODE = 200



import os
from dotenv import load_dotenv


# Evaluates whether there are environment variables that match a global variable
# If a matching environment variable is found, it returns the environment variable with the correct data type
# Otherwise, it returns the default value
def get_configuration_value(variable_name: str, default_value: any):
    if "OVERRIDE_CONFIG_IN_USE" in os.environ:
        # Verify if there's the variable exists within all env variables
        if (os.environ.get("OVERRIDE_CONFIG_IN_USE") == 'True') and variable_name in os.environ:
            env_value = os.environ[variable_name]
            # Convert env variable type from a string to a int
            if env_value.isnumeric():
                return int(env_value)
            # Convert env variable type from a string to a boolean
            elif env_value.lower() in ('true', 'false'):
                return env_value.lower() == 'true'
            else:
                return env_value
    return default_value


# loads configuration values from a specified file into the environment variables
def load_configuration_file(config_file_path: str):
    if os.path.isfile(config_file_path):
        load_dotenv(config_file_path)
        os.environ["OVERRIDE_CONFIG_IN_USE"] = 'True'
        print("Configuration file loaded successfully.", flush=True)



# Base Architecture Dashboard API Variables
AD_HTTP_PROTO = "https"
AD_API_HOST = "aimentorstudio.outsystems.com"
AD_API_ENDPOINT = "AIMentorStudioAPI/rest"
AD_API_VERSION = "1"

# Architecture Dashboard API Endpoint Variables
AD_API_SUCCESS_CODE = 200
AD_API_BAD_REQUEST_CODE = 400
AD_API_UNAUTHORIZED_CODE = 401
AD_API_FORBIDDEN_CODE = 403
AD_API_NOT_FOUND_CODE = 404
AD_API_TOO_MANY_REQ_CODE = 429
AD_API_INTERNAL_ERROR_CODE = 500

# Application specific
AD_APP_ENDPOINT = "TechnicalDebt_Application"
AD_APP_LIMIT_DEFAULT = 50
AD_APP_SUCCESS_CODE = 200

# Level specific
AD_LEVELS_ENDPOINT = "TechnicalDebt_Level"
AD_LEVELS_SUCCESS_CODE = 200

# Categories specific
AD_CATEGORIES_ENDPOINT = "TechnicalDebt_Category"
AD_CATEGORIES_SUCCESS_CODE = 200



# Base LT Variables
LIFETIME_HTTP_PROTO = "https"
LIFETIME_API_ENDPOINT = "lifetimeapi/rest"
LIFETIME_API_VERSION = 2
LIFETIME_SSL_CERT_VERIFY = True

# Applications Endpoint Variables
# Application list specific
APPLICATIONS_ENDPOINT = "applications"
APPLICATIONS_SUCCESS_CODE = 200
APPLICATIONS_EMPTY_CODE = 204
APPLICATIONS_FLAG_FAILED_CODE = 400
APPLICATIONS_FAILED_CODE = 500
# Application specific
APPLICATION_SUCCESS_CODE = 200
APPLICATION_FLAG_FAILED_CODE = 400
APPLICATION_NO_PERMISSION_CODE = 403
APPLICATION_FAILED_CODE = 404
# Application version specific
APPLICATION_VERSIONS_ENDPOINT = "versions"
APPLICATION_VERSIONS_CONTENT = "content"
APPLICATION_VERSION_SUCCESS_CODE = 200
APPLICATION_VERSIONS_EMPTY_CODE = 204
APPLICATION_VERSION_INVALID_CODE = 400
APPLICATION_VERSION_NO_PERMISSION_CODE = 403
APPLICATION_VERSION_FAILED_CODE = 404
APPLICATION_VERSION_FAILED_LIST_CODE = 500
# Application Create Version specific
APPLICATION_VERSION_CREATE_SUCCESS_CODE = 201
APPLICATION_VERSION_CREATE_INVALID_CODE = 400
APPLICATION_VERSION_CREATE_NO_PERMISSION_CODE = 403
APPLICATION_VERSION_CREATE_NO_ENVIRONMENT_CODE = 404
APPLICATION_VERSION_CREATE_FAILED_CODE = 500

# Deployments Endpoint Variables
# Deployment list specific
DEPLOYMENT_ENDPOINT = "deployment"
DEPLOYMENTS_ENDPOINT = "deployments"
DEPLOYMENTS_SUCCESS_CODE = 200
DEPLOYMENTS_EMPTY_CODE = 204
DEPLOYMENTS_INVALID_CODE = 400
DEPLOYMENTS_NO_PERMISSION_CODE = 403
DEPLOYMENTS_FAILED_CODE = 500
DEPLOYMENT_MESSAGE = "Automated deploy via OutSystems Pipeline"
# Deployment creation specific
DEPLOYMENT_PLAN_V1_API_OPS = "ApplicationVersionKeys"
DEPLOYMENT_PLAN_V2_API_OPS = "ApplicationOperations"
DEPLOYMENT_SUCCESS_CODE = 201
DEPLOYMENT_INVALID_CODE = 400
DEPLOYMENT_NO_PERMISSION_CODE = 403
DEPLOYMENT_NO_ENVIRONMENT_CODE = 404
DEPLOYMENT_FAILED_CODE = 500
# Deployment specific
DEPLOYMENT_GET_SUCCESS_CODE = 200
DEPLOYMENT_GET_NO_PERMISSION_CODE = 403
DEPLOYMENT_GET_NO_DEPLOYMENT_CODE = 404
DEPLOYMENT_GET_FAILED_CODE = 500
# Deployment delete specific
DEPLOYMENT_DELETE_SUCCESS_CODE = 204
DEPLOYMENT_DELETE_IMPOSSIBLE_CODE = 400
DEPLOYMENT_DELETE_NO_PERMISSION_CODE = 403
DEPLOYMENT_DELETE_NO_DEPLOYMENT_CODE = 404
DEPLOYMENT_DELETE_FAILED_CODE = 500
# Deployment status specific
DEPLOYMENT_STATUS_ENDPOINT = "status"
DEPLOYMENT_STATUS_SUCCESS_CODE = 200
DEPLOYMENT_STATUS_NO_PERMISSION_CODE = 403
DEPLOYMENT_STATUS_NO_DEPLOYMENT_CODE = 404
DEPLOYMENT_STATUS_FAILED_CODE = 500
# Deployment execution specific
DEPLOYMENT_START_ENDPOINT = "start"
DEPLOYMENT_ABORT_ENDPOINT = "abort"
DEPLOYMENT_CONTINUE_ENDPOINT = "continue"
DEPLOYMENT_ACTION_SUCCESS_CODE = 202
DEPLOYMENT_ACTION_IMPOSSIBLE_CODE = 400
DEPLOYMENT_ACTION_NO_PERMISSION_CODE = 403
DEPLOYMENT_ACTION_NO_DEPLOYMENT_CODE = 404
DEPLOYMENT_ACTION_FAILED_CODE = 500

# Environments Endpoint Variables
# Environment list specific
ENVIRONMENTS_ENDPOINT = "environments"
ENVIRONMENTS_SUCCESS_CODE = 200
ENVIRONMENTS_NOT_FOUND_CODE = 204
ENVIRONMENTS_FAILED_CODE = 500
# Environment application list specific
ENVIRONMENT_APPLICATIONS_ENDPOINT = "applications"
ENVIRONMENT_APP_SUCCESS_CODE = 200
ENVIRONMENT_APP_NOT_STATUS_CODE = 400
ENVIRONMENT_APP_NO_PERMISSION_CODE = 403
ENVIRONMENT_APP_NOT_FOUND = 404
ENVIRONMENT_APP_FAILED_CODE = 500
# Environment deployment zones specific
ENVIRONMENT_DEPLOYMENT_ZONES_ENDPOINT = "deploymentzones"
ENVIRONMENT_ZONES_SUCCESS_CODE = 200
ENVIRONMENT_ZONES_NOT_STATUS_CODE = 400
ENVIRONMENT_ZONES_NO_PERMISSION_CODE = 403
ENVIRONMENT_ZONES_NOT_FOUND = 404
ENVIRONMENT_ZONES_FAILED_CODE = 500
# Environment application source code specific
ENVIRONMENT_APPLICATIONS_SOURCECODE_ENDPOINT = "sourcecodeaccess"
ENVIRONMENT_SOURCECODE_LINK_SUCCESS_CODE = 200
ENVIRONMENT_SOURCECODE_PACKAGE_SUCCESS_CODE = 201
ENVIRONMENT_SOURCECODE_FAILED_CODE = 500
# Solutions specific
ENVIRONMENT_SOLUTION_ENDPOINT = "solution"
ENVIRONMENT_SOLUTION_SUCCESS_CODE = 200
ENVIRONMENT_SOLUTION_NOT_STATUS_CODE = 400
ENVIRONMENT_SOLUTION_NO_PERMISSION_CODE = 403
ENVIRONMENT_SOLUTION_NOT_FOUND = 404
ENVIRONMENT_SOLUTION_FAILED_CODE = 500
# Solutions status specific
ENVIRONMENT_SOLUTION_STATUS_ENDPOINT = "solutionstatus"
ENVIRONMENT_SOLUTION_STATUS_SUCCESS_CODE = 200
ENVIRONMENT_SOLUTION_STATUS_NOT_STATUS_CODE = 400
ENVIRONMENT_SOLUTION_STATUS_NO_PERMISSION_CODE = 403
ENVIRONMENT_SOLUTION_STATUS_NOT_FOUND = 404
ENVIRONMENT_SOLUTION_STATUS_FAILED_CODE = 500
# Solutions link specific
ENVIRONMENT_SOLUTION_LINK_SUCCESS_CODE = 200
ENVIRONMENT_SOLUTION_LINK_FAILED_CODE = 400

# Downloads Endpoint Variables
DOWNLOADS_ENDPOINT = "downloads"
DOWNLOAD_SUCCESS_CODE = 200
DOWNLOAD_INVALID_KEY_CODE = 400
DOWNLOAD_NO_PERMISSION_CODE = 403
DOWNLOAD_NOT_FOUND = 404
DOWNLOAD_FAILED_CODE = 500



# Python Modules
import requests

# Custom Modules
from outsystems.exceptions.invalid_json_response import InvalidJsonResponseError
from outsystems.vars.properties_vars import PROPERTIES_API_HTTP_PROTO, PROPERTIES_API_ENDPOINT, PROPERTIES_API_VERSION, PROPERTIES_API_SSL_CERT_VERIFY
# Functions
from outsystems.vars.vars_base import get_configuration_value


# Method that builds the Properties API endpoint based on the environment host
def build_properties_api_url(properties_http_proto: str, lt_url: str, properties_api_endpoint: str, properties_api_version: int):
    return "{}://{}/{}/v{}".format(properties_http_proto, lt_url, properties_api_endpoint, properties_api_version)


# Sends a PUT request to Properties API, with a payload. The json part is ignored
def send_properties_put_request(lt_url: str, token: str, api_endpoint: str, payload: str):
    # Auth token + content type json
    headers = {'content-type': 'application/json',
               'authorization': 'Bearer ' + token}
    # Format the request URL to include the api endpoint
    properties_api_url = build_properties_api_url(PROPERTIES_API_HTTP_PROTO, lt_url, PROPERTIES_API_ENDPOINT, PROPERTIES_API_VERSION)
    request_string = "{}/{}".format(properties_api_url, api_endpoint)
    response = requests.put(
        request_string, data=payload, json=None, headers=headers, verify=get_configuration_value("PROPERTIES_API_SSL_CERT_VERIFY", PROPERTIES_API_SSL_CERT_VERIFY))
    response_obj = {"http_status": response.status_code, "response": {}}
    if len(response.text) > 0:
        try:
            response_obj["response"] = response.json()
        except:
            raise InvalidJsonResponseError(
                "PUT {}: The JSON response could not be parsed. Response: {}".format(request_string, response.text))

    return response_obj



# Python Modules

# Custom Modules
# Exceptions
from outsystems.exceptions.environment_not_found import EnvironmentNotFoundError
# Variables
from outsystems.vars.manifest_vars import MANIFEST_ENVIRONMENT_KEY, MANIFEST_ENVIRONMENT_NAME, MANIFEST_ENVIRONMENT_LABEL, \
    MANIFEST_ENVIRONMENT_DEFINITIONS, MANIFEST_CONFIGURATION_ITEMS, MANIFEST_CONFIG_ITEM_VALUES, MANIFEST_MODULE_KEY, MANIFEST_MODULE_NAME, \
    MANIFEST_CONFIG_ITEM_KEY, MANIFEST_CONFIG_ITEM_NAME, MANIFEST_CONFIG_ITEM_TYPE, MANIFEST_CONFIG_ITEM_TARGET_VALUE, MANIFEST_DEPLOYMENT_NOTES
from outsystems.vars.lifetime_vars import DEPLOYMENT_MESSAGE


# Returns the environment details: tuple(Name, Key)
def get_environment_details(manifest: dict, environment_label: str):
    environment_definition = next(filter(lambda x: x[MANIFEST_ENVIRONMENT_LABEL] == environment_label, manifest[MANIFEST_ENVIRONMENT_DEFINITIONS]), None)
    if environment_definition:
        return (environment_definition[MANIFEST_ENVIRONMENT_NAME], environment_definition[MANIFEST_ENVIRONMENT_KEY])
    else:
        raise EnvironmentNotFoundError(
            "Failed to retrieve the environment key from label. Please make sure the label is correct. Environment label: {}".format(environment_label))


# Returns the configuration items for the target environment key
def get_configuration_items_for_environment(manifest: dict, target_env_key: str):
    config_items = []
    if MANIFEST_CONFIGURATION_ITEMS in manifest:
        for cfg_item in manifest[MANIFEST_CONFIGURATION_ITEMS]:
            target_value = next(filter(lambda x: x[MANIFEST_ENVIRONMENT_KEY] == target_env_key, cfg_item[MANIFEST_CONFIG_ITEM_VALUES]), None)
            if target_value:
                # Add it to the config items list
                config_items.append({
                    MANIFEST_MODULE_KEY: cfg_item[MANIFEST_MODULE_KEY],
                    MANIFEST_MODULE_NAME: cfg_item[MANIFEST_MODULE_NAME],
                    MANIFEST_CONFIG_ITEM_KEY: cfg_item[MANIFEST_CONFIG_ITEM_KEY],
                    MANIFEST_CONFIG_ITEM_NAME: cfg_item[MANIFEST_CONFIG_ITEM_NAME],
                    MANIFEST_CONFIG_ITEM_TYPE: cfg_item[MANIFEST_CONFIG_ITEM_TYPE],
                    MANIFEST_CONFIG_ITEM_TARGET_VALUE: target_value[MANIFEST_CONFIG_ITEM_TARGET_VALUE],
                    MANIFEST_ENVIRONMENT_NAME: target_value[MANIFEST_ENVIRONMENT_NAME]
                })

    return config_items


# Returns the deployment notes
def get_deployment_notes(manifest: dict):
    if MANIFEST_DEPLOYMENT_NOTES in manifest:
        return manifest[MANIFEST_DEPLOYMENT_NOTES]
    else:
        return DEPLOYMENT_MESSAGE



# Base CICD Probe Variables
PROBE_HTTP_PROTO = "https"
PROBE_API_ENDPOINT = "CI_CDProbe/rest"
PROBE_API_VERSION = 1
PROBE_API_SSL_CERT_VERIFY = True

# Scan Endpoints
SCAN_BDD_TESTS_ENDPOINT = "ScanBDDTestEndpoints"
PROBE_SCAN_SUCCESS_CODE = 200

# Application Dependencies Endpoint
GET_APPLICATION_DEPENDENCIES_ENDPOINT = "GetApplicationDependencies"
PROBE_DEPENDENCIES_SUCCESS_CODE = 200



# Python Modules
import sys
import os
import argparse
from packaging.version import Version

# Workaround for Jenkins:
# Set the path to include the outsystems module
# Jenkins exposes the workspace directory through env.
if "WORKSPACE" in os.environ:
    sys.path.append(os.environ['WORKSPACE'])
else:  # Else just add the project dir
    sys.path.append(os.getcwd())

# Custom Modules
# Variables
from outsystems.vars.file_vars import ARTIFACT_FOLDER
from outsystems.vars.lifetime_vars import LIFETIME_HTTP_PROTO, LIFETIME_API_ENDPOINT, LIFETIME_API_VERSION

# Functions
from outsystems.file_helpers.file import load_data
from outsystems.lifetime.lifetime_environments import get_environment_key
from outsystems.lifetime.lifetime_base import build_lt_endpoint
from outsystems.lifetime.lifetime_applications import set_application_version, get_running_app_version
from outsystems.vars.vars_base import load_configuration_file
# Exceptions
from outsystems.exceptions.invalid_parameters import InvalidParametersError


# ############################################################# SCRIPT ##############################################################
def valid_tag_number(artifact_dir: str, lt_endpoint: str, lt_token: str, env_name: str, env_key: str, app: dict):
    # Get the app running version on the source environment. It will only retrieve tagged applications
    running_app = get_running_app_version(artifact_dir, lt_endpoint, lt_token, env_key, app_name=app["ApplicationName"])

    if Version(running_app["Version"]) < Version(app["VersionNumber"]):
        return True

    print("Skipping tag! Application '{}' current tag ({}) on {} is greater than or equal to the manifest data ({}). ".format(app["ApplicationName"], running_app["Version"], env_name, app["VersionNumber"]), flush=True)
    return False


def main(artifact_dir: str, lt_http_proto: str, lt_url: str, lt_api_endpoint: str, lt_api_version: int, lt_token: str, dest_env: str, app_list: list, dep_manifest: list, trigger_manifest: dict, include_test_apps: bool):
    # Builds the LifeTime endpoint
    lt_endpoint = build_lt_endpoint(lt_http_proto, lt_url, lt_api_endpoint, lt_api_version)
    # Get the environment keys
    dest_env_key = get_environment_key(artifact_dir, lt_endpoint, lt_token, dest_env)

    # the app versions MUST come from that a file
    # either deployment or trigger manifest file
    if dep_manifest:
        for deployed_app in dep_manifest:
            if deployed_app["ApplicationName"] in app_list:
                set_application_version(lt_endpoint, lt_token, dest_env_key, deployed_app["ApplicationKey"], deployed_app["ChangeLog"], deployed_app["Version"], None)
                print("{} application successuflly tagged as {} on {}".format(deployed_app["ApplicationName"], deployed_app["Version"], dest_env), flush=True)
    elif trigger_manifest:
        for deployed_app in trigger_manifest["ApplicationVersions"]:
            if not deployed_app["IsTestApplication"] or (deployed_app["IsTestApplication"] and include_test_apps):
                if valid_tag_number(artifact_dir, lt_endpoint, lt_token, dest_env, dest_env_key, deployed_app):
                    set_application_version(lt_endpoint, lt_token, dest_env_key, deployed_app["ApplicationKey"], deployed_app["ChangeLog"], deployed_app["VersionNumber"], None)
                    print("{} application successuflly tagged as {} on {}".format(deployed_app["ApplicationName"], deployed_app["VersionNumber"], dest_env), flush=True)
                else:
                    continue

# End of main()


if __name__ == "__main__":
    # Argument menu / parsing
    parser = argparse.ArgumentParser()
    parser.add_argument("-a", "--artifacts", type=str, default=ARTIFACT_FOLDER,
                        help="Name of the artifacts folder. Default: \"Artifacts\"")
    parser.add_argument("-u", "--lt_url", type=str, required=True,
                        help="URL for LifeTime environment, without the API endpoint. Example: \"https://<lifetime_host>\"")
    parser.add_argument("-t", "--lt_token", type=str, required=True,
                        help="Token for LifeTime API calls.")
    parser.add_argument("-v", "--lt_api_version", type=int, default=LIFETIME_API_VERSION,
                        help="LifeTime API version number. If version <= 10, use 1, if version >= 11, use 2. Default: 2")
    parser.add_argument("-e", "--lt_endpoint", type=str, default=LIFETIME_API_ENDPOINT,
                        help="(Optional) Used to set the API endpoint for LifeTime, without the version. Default: \"lifetimeapi/rest\"")
    parser.add_argument("-d", "--destination_env", type=str, required=True,
                        help="Name, as displayed in LifeTime, of the destination environment where you want to deploy the apps. (if in Airgap mode should be the hostname of the destination environment where you want to deploy the apps)")
    parser.add_argument("-l", "--app_list", type=str,
                        help="(Optional) Comma separated list of apps you want to tag. Example: \"App1,App2 With Spaces,App3_With_Underscores\"")
    parser.add_argument("-f", "--manifest_file", type=str, required=True,
                        help="Manifest file path (either deployment or trigger).")
    parser.add_argument("-i", "--include_test_apps", action='store_true',
                        help="(Optional) Flag that indicates if applications marked as \"Test Application\" in the trigger manifest are included for tagging.")
    parser.add_argument("-cf", "--config_file", type=str,
                        help="Config file path. Contains configuration values to override the default ones.")

    args = parser.parse_args()

    # Load config file if exists
    if args.config_file:
        load_configuration_file(args.config_file)
    # Parse the artifact directory
    artifact_dir = args.artifacts
    # Parse the API endpoint
    lt_api_endpoint = args.lt_endpoint
    # Parse the LT Url and split the LT hostname from the HTTP protocol
    # Assumes the default HTTP protocol = https
    lt_http_proto = LIFETIME_HTTP_PROTO
    lt_url = args.lt_url
    if lt_url.startswith("http://"):
        lt_http_proto = "http"
        lt_url = lt_url.replace("http://", "")
    else:
        lt_url = lt_url.replace("https://", "")
    if lt_url.endswith("/"):
        lt_url = lt_url[:-1]
    # Parte LT API Version
    lt_version = args.lt_api_version
    # Parse the LT Token
    lt_token = args.lt_token
    # Parse Destination Environment
    dest_env = args.destination_env

    # Parse Manifest file if it exists
    # Based on the file content it can be a deployment manifest (list-based) or trigger manifest (dict-based)
    manifest_file = None
    if args.manifest_file:
        manifest_path = os.path.split(args.manifest_file)
        manifest_file = load_data(manifest_path[0], manifest_path[1])

    dep_manifest = manifest_file if type(manifest_file) is list else None
    trigger_manifest = manifest_file if type(manifest_file) is dict else None

    if dep_manifest and not args.app_list:
        raise InvalidParametersError("--app_list parameter is required for Deployment Manifest operation")

    # Parse App list
    apps = None
    if args.app_list:
        _apps = args.app_list
        apps = _apps.split(',')

    # Parse Include Test Apps flag
    include_test_apps = args.include_test_apps
    # Calls the main script
    main(artifact_dir, lt_http_proto, lt_url, lt_api_endpoint, lt_version, lt_token, dest_env, apps, dep_manifest, trigger_manifest, include_test_apps)  # type: ignore
